from . import TestCase
from .models import ModelWithComputedCollationField


class ComputedCollationFieldTests(TestCase):
    """Tests for `ComputedCollationField`."""

    def test_model(self):
        """Tests for a model using a `ComputedCollationField`."""
        ModelWithComputedCollationField.objects.create(name="demo1")



import google

from . import TestCase
from .models import NullableFieldModel, StringPkModel


class QueryByKeysTest(TestCase):
    """
        Tests for the Get optimisation when keys are
        included in all branches of a query.
    """

    databases = "__all__"

    def test_missing_results_are_skipped(self):
        NullableFieldModel.objects.create(pk=1)
        NullableFieldModel.objects.create(pk=5)

        results = NullableFieldModel.objects.filter(
            pk__in=[1, 2, 3, 4, 5]
        ).order_by("nullable").values_list("pk", flat=True)

        self.assertCountEqual(results, [1, 5])

    def test_none_namespace(self):
        NullableFieldModel.objects.using("nonamespace").create(pk=1)
        NullableFieldModel.objects.using("nonamespace").create(pk=5)

        results = NullableFieldModel.objects.using(
            "nonamespace").filter(
                pk__in=[1, 2, 3, 4, 5]
        ).order_by("nullable").values_list("pk", flat=True)

        self.assertCountEqual(results, [1, 5])

    def test_large_number_of_keys(self):
        keys = []

        for i in range(1001):
            keys.append(NullableFieldModel.objects.create(pk=i + 1).pk)

        try:
            results = list(NullableFieldModel.objects.filter(pk__in=keys))
        except google.api_core.exceptions.InvalidArgument:
            self.fail("Didn't correctly deal with a large number of keys")

        self.assertEqual(len(results), 1001)
        self.assertCountEqual([x.pk for x in results], keys)

    def test_multiple_pk_filters(self):
        for i in range(10):
            NullableFieldModel.objects.create(pk=i + 1)

        qs = NullableFieldModel.objects.all()
        self.assertEqual(qs.count(), 10)

        qs = qs.filter(pk__lt=10)
        self.assertEqual(qs.count(), 9)

        qs = qs.filter(pk__gte=2)
        self.assertEqual(qs.count(), 8)

        qs = qs.filter(pk__gte=3)
        self.assertEqual(qs.count(), 7)

    def test_multiple_str_pk_filters(self):
        for i in range(9):
            StringPkModel.objects.create(pk=str(i + 1))

        qs = StringPkModel.objects.all()
        self.assertEqual(qs.count(), 9)

        qs = qs.filter(pk__lt=str(9))
        self.assertEqual(qs.count(), 8)

        qs = qs.filter(pk__gte=str(2))
        self.assertEqual(qs.count(), 7)

        qs = qs.filter(pk__gte=str(3))
        self.assertEqual(qs.count(), 6)



# STANDARD LIBRARY
import logging

# 3RD PARTY
from django.db.backends.base.schema import BaseDatabaseSchemaEditor

# DJANGAE
from gcloudc.db.backends.datastore.dbapi import CouldBeSupportedError

logger = logging.getLogger("djangae")


class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
    """
    Work in progress!
    """

    def execute(self, sql, params=[]):
        """ Rather than executing SQL, we can probably (hopefully) do something cunning here
            involving deferring tasks to go and perform operations on the DB, probably using
            defer_iteration, or some kind of extension of that.

            The key will be knowing when the tasks have finished running.

            When running locally, there won't be a server running or a task queue stub set up
            (assuming we're doing mangage.py migrate) and so we will want to patch deferred.defer
            so that instead of being deferred to tasks, the functions are called immediately.
        """
        raise NotImplementedError("See docstring")

    def column_sql(self, model, field, include_default=False):
        """
        Takes a field and returns its column definition.
        The field must already have had set_attributes_from_name called.
        """
        # Erm... is this part of the external interface that we have to implement?  Hopefully not.

    def quote_value(self, value):
        """ SQL, lolz. """
        return value

    def create_model(self, model):
        """ Responsible for creating the table, indexes and unique constraints.
            I'm pretty sure that we can skip all of this for the Datastore.
        """
        pass

    def delete_model(self, model):
        """ Deletes a model from the database. """
        # The plan:
        # 1. Delete unique markers
        # 2. Delete entities
        # 3. Profit.
        raise NotImplementedError("See comments.")

    def alter_unique_together(self, model, old_unique_together, new_unique_together):
        """ Deals with a model changing its unique_together. """
        # The plan:
        # 1. Delete the old unique markers
        # 2. Create the new unique markers
        # Hopefully we can use djangae.contrib.uniquetool for this?
        raise NotImplementedError("See comments.")

    def alter_index_together(self, model, old_index_together, new_index_together):
        """ Deals with a model changing its index_together.
            Irrelevant for the Datastore.
        """
        pass

    def alter_db_table(self, model, old_db_table, new_db_table):
        """ Renames the table a model points to. """
        # Ideally this would be done as 3 separate steps:
        # (1) copy entities to new Datastore Kind
        # (2) deploy code which references new "table" (Kind)
        # (3) delete old entities
        # But Django's `manage.py migrate` approach isn't going to let us do that.
        # So we have 2 options:
        # 1. Implement it as a single mapreduce thing which *moves* the entities in a single
        # operation, thereby forcing your site to be offline/broken for a while.
        # 2. Not implement it and raise AreYouOffYourRockerError.
        raise NotImplementedError("See comments.")

    def alter_db_tablespace(self, model, old_db_tablespace, new_db_tablespace):
        """ Moves a model's table between tablespaces. """
        # Say what?
        # I guess this is equivalent to the Datastore namespaces.
        raise CouldBeSupportedError()

    def add_field(self, model, field):
        """ Creates a field on a model. """
        # Althought the Datastore is schemaless, when adding a new field we still want to deal with:
        # 1. If the default is not None, populate existing entities with default value.
        # 2. If the default is None, populate existing entities with None so that they are indexed.

        # Basically, there's no need to do any of this, *unless* you want to filter on this field.
        # But we don't know whether the project needs to filter on the field or not, so we have 2 options:
        # 1. Just run the populate tasks anyway.
        # 2. Find somewhere to add some kind of yes_please_populate_existing_entities option... ?
        # 3. Don't do anything, and allow the project to just specify the population task manually
        #    using the RunPython or RunDatastoreThing functionality that we will provide.

        # @adamalton votes for #3!

        # Oh, and did I mention that we may need to add unique markers if the field is unique?
        raise NotImplementedError

    def remove_field(self, model, field):
        """ Removes a field from a model. """
        # We need to:
        # 1. Delete any unique markers if the field was unique.
        # 2. Delete the data for the field.  Here we have several options:
        #     a. Not bother because deleting it will be almost as expensive as storing it (short term).
        #     b. Delete it.
        #     c. Allow the project to use our RunDatastoreThing functionality to delete it if they want to.
        raise NotImplementedError

    def alter_field(self, model, old_field, new_field, strict=False):
        """ Allows a field's type, uniqueness, nullability, default, column, constraints etc. to be modified. """
        # Nullability: doesn't matter.
        # Uniqueness: yes, we'll need to add/remove unique markers.
        # Default: doesn't matter.
        # Column: Erm... see `alter_db_table`.
        raise NotImplementedError



import datetime
import json
import logging
import re
from itertools import chain

from django.db import (
    NotSupportedError,
    connections,
)
from django.db.models import AutoField
from django.core.exceptions import EmptyResultSet

from . import POLYMODEL_CLASS_ATTRIBUTE
from .indexing import (
    add_special_index,
    get_indexer,
)
from .utils import (
    ensure_datetime,
    get_field_from_column,
    get_top_concrete_parent,
    has_concrete_parents,
)

logger = logging.getLogger(__name__)


VALID_QUERY_KINDS = ("SELECT", "UPDATE", "INSERT", "DELETE", "COUNT", "AVERAGE")

VALID_ANNOTATIONS = {"MIN": min, "MAX": max, "SUM": sum, "COUNT": len, "AVG": lambda x: (sum(x) / len(x))}

VALID_CONNECTORS = ("AND", "OR")


VALID_OPERATORS = ("=", "<", ">", "<=", ">=", "IN")


def convert_operator(operator):
    if operator == "exact":
        return "="
    elif operator == "gt":
        return ">"
    elif operator == "lt":
        return "<"
    elif operator == "gte":
        return ">="
    elif operator == "lte":
        return "<="

    return operator.upper()


class WhereNode(object):
    def __init__(self, using):
        self.using = using

        self.column = None
        self.operator = None
        self.value = None
        self.output_field = None
        self.will_never_return_results = False
        self.lookup_name = None

        self.children = []
        self.connector = "AND"
        self.negated = False

    @property
    def is_leaf(self):
        return bool(self.column and self.operator)

    def set_connector(self, connector):
        self.connector = connector

    def append_child(self, node):
        self.children.append(node)

    def set_leaf(self, column, operator, value, is_pk_field, negated, lookup_name, namespace, target_field=None):

        # We need access to the Datastore client to access the Key factory
        wrapper = connections[self.using]
        wrapper.ensure_connection()
        gclient = wrapper.connection.gclient

        assert column
        assert operator
        assert isinstance(is_pk_field, bool)
        assert isinstance(negated, bool)

        if operator == "iexact" and isinstance(target_field, AutoField):
            # When new instance is created, automatic primary key 'id'
            # does not generate '_idx_iexact_id'.
            # As the primary key 'id' (AutoField) is integer and is always case insensitive,
            # we can deal with 'id_iexact=' query by using 'exact' rather than 'iexact'.
            operator = "exact"
            value = int(value)

        if is_pk_field:
            # If this is a primary key, we need to make sure that the value
            # we pass to the query is a datastore Key. We have to deal with IN queries here
            # because they aren't flattened until the DNF stage
            model = get_top_concrete_parent(target_field.model)
            table = model._meta.db_table

            if isinstance(value, (list, tuple)):
                value = [gclient.key(table, x, namespace=namespace) for x in value if x]
            else:
                # Django 1.11 has operators as symbols, earlier versions use "exact" etc.
                if (operator == "isnull" and value is True) or (
                    operator in ("exact", "lt", "lte", "<", "<=", "=") and not value
                ):
                    # id=None will never return anything and
                    # Empty strings and 0 are forbidden as keys
                    self.will_never_return_results = True
                elif operator in ("gt", "gte", ">", ">=") and not value:
                    # If the value is 0 or "", then we need to manipulate the value and operator here to
                    # get the right result (given that both are invalid keys) so for both we return
                    # >= 1 or >= "\0" for strings
                    if isinstance(value, int):
                        value = 1
                    else:
                        value = "\0"

                    value = gclient.key(table, value, namespace=namespace)
                    operator = "gte"
                else:
                    value = gclient.key(table, value, namespace=namespace)
            column = "__key__"

        # Do any special index conversions necessary to perform this lookup
        special_indexer = get_indexer(target_field, operator)

        if special_indexer:
            if is_pk_field:
                column = model._meta.pk.column
                value = str(value.id_or_name)

            add_special_index(connections[self.using], target_field.model, column, special_indexer, operator, value)
            index_type = special_indexer.prepare_index_type(operator, value)
            value = special_indexer.prep_value_for_query(
                value, model=target_field.model, column=column, connection=connections[self.using]
            )
            column = special_indexer.indexed_column_name(column, value, index_type)
            operator = special_indexer.prep_query_operator(operator)

        self.column = column
        self.operator = convert_operator(operator)
        self.value = value
        self.lookup_name = lookup_name

    def __iter__(self):
        for child in chain(*map(iter, self.children)):
            yield child
        yield self

    def __repr__(self):
        if self.is_leaf:
            return "[%s%s%s]" % (self.column, self.operator, self.value)
        else:
            return "(%s:%s%s)" % (
                self.connector,
                "!" if self.negated else "",
                ",".join([repr(x) for x in self.children]),
            )

    def __eq__(self, rhs):
        if self.is_leaf != rhs.is_leaf:
            return False

        if self.is_leaf:
            return self.column == rhs.column and self.value == rhs.value and self.operator == rhs.operator
        else:
            return self.connector == rhs.connector and self.children == rhs.children

    def __hash__(self):
        if self.is_leaf:
            return hash((self.column, self.value, self.operator))
        else:
            return hash((self.connector,) + tuple([hash(x) for x in self.children]))


class Query(object):
    def __init__(self, model, kind):
        assert kind in VALID_QUERY_KINDS

        self.model = model
        self.concrete_model = get_top_concrete_parent(model)
        self.kind = kind

        self.projection_possible = True
        self.tables = []

        self.columns = None  # None means all fields
        self.init_list = []

        self.distinct = False
        self.order_by = []
        self.row_data = []  # For insert/updates
        self._where = None
        self.low_mark = self.high_mark = None

        self.annotations = []
        self.per_entity_annotations = []
        self.extra_selects = []
        self.polymodel_filter_added = False

        # A list of PKs that should be excluded from the resultset
        self.excluded_pks = set()

    @property
    def is_normalized(self):
        """
            Returns True if this query has a normalized where tree
        """
        if not self.where:
            return True

        # Only a leaf node, return True
        if not self.where.is_leaf:
            return True

        # If we have children, and they are all leaf nodes then this is a normalized
        # query
        return self.where.connector == "OR" and self.where.children and all(x.is_leaf for x in self.where.children)

    def add_extra_select(self, column, lookup):
        if lookup.lower().startswith("select "):
            raise ValueError("SQL statements aren't supported with extra(select=)")

        # Boolean expression test
        bool_expr = r"(?P<lhs>[a-zA-Z0-9_]+)\s?(?P<op>[=|>|<]{1,2})\s?(?P<rhs>[\w+|']+)"

        # Operator expression test
        op_expr = r"(?P<lhs>[a-zA-Z0-9_]+)\s?(?P<op>[+|-|/|*])\s?(?P<rhs>[\w+|']+)"

        OP_LOOKUP = {
            "=": lambda x, y: x == y,
            "is": lambda x, y: x == y,
            "<": lambda x, y: x < y,
            ">": lambda x, y: x > y,
            ">=": lambda x, y: x >= y,
            "<=": lambda x, y: x <= y,
            "+": lambda x, y: x + y,
            "-": lambda x, y: x - y,
            "/": lambda x, y: x / y,
            "*": lambda x, y: x * y,
        }

        for regex in (bool_expr, op_expr):
            match = re.match(regex, lookup)
            if match:
                lhs = match.group("lhs")
                rhs = match.group("rhs")
                op = match.group("op").lower()
                if op in OP_LOOKUP:
                    self.extra_selects.append((column, (OP_LOOKUP[op], (lhs, rhs))))
                else:
                    raise ValueError("Unsupported operator")
                return

        # Assume literal
        self.extra_selects.append((column, (lambda x: x, [lookup])))

    def add_source_table(self, table):
        if table in self.tables:
            return

        self.tables.append(table)

    def set_distinct(self, distinct_fields):
        self.distinct = True
        if distinct_fields:
            for field in distinct_fields:
                self.add_projected_column(field)
        elif not self.columns:
            for field in self.model._meta.fields:
                self.add_projected_column(field.column)

    def add_order_by(self, column):
        self.order_by.append(column)

    def add_annotation(self, column, annotation):
        # The Trunc annotation class doesn't exist in Django 1.8, hence we compare by
        # strings, rather than importing the class to compare it
        name = annotation.__class__.__name__
        if name == "Count":
            return  # Handled elsewhere

        if name not in ("Trunc", "Col", "Date", "DateTime"):
            raise NotSupportedError("Unsupported annotation %s" % name)

        def process_date(value, lookup_type):
            value = ensure_datetime(value)
            ret = datetime.datetime.utcfromtimestamp(0)

            POSSIBLE_LOOKUPS = ("year", "month", "day", "hour", "minute", "second")

            ret = ret.replace(
                value.year,
                value.month if lookup_type in POSSIBLE_LOOKUPS[1:] else ret.month,
                value.day if lookup_type in POSSIBLE_LOOKUPS[2:] else ret.day,
                value.hour if lookup_type in POSSIBLE_LOOKUPS[3:] else ret.hour,
                value.minute if lookup_type in POSSIBLE_LOOKUPS[4:] else ret.minute,
                value.second if lookup_type in POSSIBLE_LOOKUPS[5:] else ret.second,
            )

            return ret

        # Abuse the extra_select functionality
        if name == "Col":
            self.extra_selects.append((column, (lambda x: x, [column])))
        elif name in ("Trunc", "Date", "DateTime"):
            # Trunc stores the source column and the lookup type differently to Date
            # which is why we have the getattr craziness here
            lookup_column = (
                annotation.lhs.output_field.column if name == "Trunc" else getattr(annotation, "lookup", column)
            )

            lookup_type = getattr(annotation, "lookup_type", getattr(annotation, "kind", None))
            assert lookup_type

            self.extra_selects.append((column, (lambda x: process_date(x, lookup_type), [lookup_column])))
            # Override the projection so that we only get this column
            self.columns = set([lookup_column])

    def add_projected_column(self, column):
        self.init_list.append(column)

        if not self.projection_possible:
            # If we previously tried to add a column that couldn't be
            # projected, we don't try and add any more
            return

        field = get_field_from_column(self.model, column)

        if field is None:
            raise NotSupportedError(
                "{} is not a valid column for the queried model. Did you try to join?".format(column)
            )

        if field.db_type(self.connection) in ("bytes", "text", "list", "set"):
            logger.warn("Disabling projection query as %s is an unprojectable type", column)
            self.columns = None
            self.projection_possible = False
            return

        if not self.columns:
            self.columns = set([column])
        else:
            self.columns.add(column)

    def add_row(self, data):
        assert self.columns
        assert len(data) == len(self.columns)

        self.row_data.append(data)

    def prepare(self):
        if not self.init_list:
            self.init_list = [x.column for x in self.model._meta.fields]

        self._remove_impossible_branches()
        self._remove_erroneous_isnull()
        self._remove_negated_empty_in()
        self._add_inheritence_filter()
        self._populate_excluded_pks()
        self._disable_projection_if_fields_used_in_equality_filter()
        self._check_only_single_inequality_filter()

    @property
    def where(self):
        return self._where

    @where.setter
    def where(self, where):
        assert where is None or isinstance(where, WhereNode)

        self._where = where

    def _populate_excluded_pks(self):
        if not self._where:
            return

        self.excluded_pks = set()

        def walk(node, negated):
            if node.connector == "OR":
                # We can only process AND nodes, if we hit an OR we can't
                # use the excluded PK optimization
                return

            if node.negated:
                negated = not negated

            for child in node.children[:]:
                # As more than one inequality filter is not allowed on the datastore
                # this leaf + count check is probably pointless, but at least if you
                # do try to exclude two things it will blow up in the right place and not
                # return incorrect results
                if child.is_leaf and len(node.children) == 1:
                    if negated and child.operator == "=" and child.column == "__key__":
                        self.excluded_pks.add(child.value)
                        node.children.remove(child)
                    elif negated and child.operator == "IN" and child.column == "__key__":
                        [self.excluded_pks.add(x) for x in child.value]
                        node.children.remove(child)
                else:
                    walk(child, negated)

            node.children = [x for x in node.children if x.children or x.column]

        walk(self._where, False)

        if not self._where.children:
            self._where = None

    def _remove_negated_empty_in(self):
        """
            An empty exclude(id__in=[]) is pointless, but can cause trouble
            during denormalization. We remove such nodes here.
        """
        if not self._where:
            return

        def walk(node, negated):
            if node.negated:
                negated = node.negated

            for child in node.children[:]:
                if negated and child.operator == "IN" and not child.value:
                    node.children.remove(child)

                walk(child, negated)

            node.children = [x for x in node.children if x.children or x.column]

        had_where = bool(self._where.children)
        walk(self._where, False)

        # Reset the where if that was the only filter
        if had_where and not bool(self._where.children):
            self._where = None

    def _remove_erroneous_isnull(self):
        # This is a little crazy, but bear with me...
        # If you run a query like this:  filter(thing=1).exclude(field1="test") where field1 is
        # null-able you'll end up with a negated branch in the where tree which is:

        #           AND (negated)
        #          /            \
        #   field1="test"   field1__isnull=False

        # This is because on SQL, field1 != "test" won't give back rows where field1 is null, so
        # django has to include the negated isnull=False as well in order to get back the null rows
        # as well.  On App Engine though None is just a value, not the lack of a value, so it's
        # enough to just have the first branch in the negated node and in fact, if you try to use
        # the above tree, it will result in looking for:
        #  field1 < "test" and field1 > "test" and field1__isnull=True
        # which returns the wrong result (it'll return when field1 == None only)

        def walk(node, negated):
            if node.negated:
                negated = not negated

            if not node.is_leaf:
                equality_fields = set()
                negated_isnull_fields = set()
                isnull_lookup = {}

                for child in node.children[:]:
                    if negated:
                        if child.lookup_name != "isnull":
                            equality_fields.add(child.column)
                            if child.column in negated_isnull_fields:
                                node.children.remove(isnull_lookup[child.column])
                        else:
                            negated_isnull_fields.add(child.column)
                            if child.column in equality_fields:
                                node.children.remove(child)
                            else:
                                isnull_lookup[child.column] = child

                    walk(child, negated)

        if self.where:
            walk(self._where, False)

    def _remove_impossible_branches(self):
        """
            If we mark a child node as never returning results we either need to
            remove those nodes, or remove the branches of the tree they are on depending
            on the connector of the parent node.
        """
        if not self._where:
            return

        def walk(node, negated):
            if node.negated:
                negated = not negated

            for child in node.children[:]:
                walk(child, negated)

                if child.will_never_return_results:
                    if node.connector == "AND":
                        if child.negated:
                            node.children.remove(child)
                        else:
                            node.will_never_return_results = True
                    else:
                        # OR
                        if not child.negated:
                            node.children.remove(child)
                            if not node.children:
                                node.will_never_return_results = True
                        else:
                            node.children[:] = []

        walk(self._where, False)

        if self._where.will_never_return_results:
            # We removed all the children of the root where node, so no results
            raise EmptyResultSet()

    def _check_only_single_inequality_filter(self):
        inequality_fields = set()

        def walk(node, negated):
            if node.negated:
                negated = not negated

            for child in node.children[:]:
                if (negated and child.operator == "=") or child.operator in (">", "<", ">=", "<="):
                    inequality_fields.add(child.column)
                walk(child, negated)

            if len(inequality_fields) > 1:
                raise NotSupportedError(
                    "You can only have one inequality filter per query on the rpc. "
                    "Filters were: %s" % " ".join(inequality_fields)
                )

        if self.where:
            walk(self._where, False)

    def _disable_projection_if_fields_used_in_equality_filter(self):
        if not self._where or not self.columns:
            return

        equality_columns = set()

        def walk(node):
            if not node.is_leaf:
                for child in node.children:
                    walk(child)
            elif node.operator == "=" or node.operator == "IN":
                equality_columns.add(node.column)

        walk(self._where)

        if equality_columns and equality_columns.intersection(self.columns):
            self.columns = None
            self.projection_possible = False

    def _add_inheritence_filter(self):
        """
            We support inheritence with polymodels. Whenever we set
            the 'where' on this query, we manipulate the tree so that
            the lookups are ANDed with a filter on 'class = db_table'
            and on inserts, we add the 'class' column if the model is part
            of an inheritance tree.

            We only do any of this if the model has concrete parents and isn't
            a proxy model
        """
        if has_concrete_parents(self.model) and not self.model._meta.proxy:
            if self.polymodel_filter_added:
                return

            new_filter = WhereNode(self.connection.alias)
            new_filter.column = POLYMODEL_CLASS_ATTRIBUTE
            new_filter.operator = "="
            new_filter.value = self.model._meta.db_table

            # We add this bare AND just to stay consistent with what Django does
            new_and = WhereNode(self.connection.alias)
            new_and.connector = "AND"
            new_and.children = [new_filter]

            new_root = WhereNode(self.connection.alias)
            new_root.connector = "AND"
            new_root.children = [new_and]
            if self._where:
                # Add the original where if there was one
                new_root.children.append(self._where)
            self._where = new_root

            self.polymodel_filter_added = True

    def serialize(self):
        """
            The idea behind this function is to provide a way to serialize this
            query to a string which can be compared to another query. Pickle won't
            work as some of the values etc. might not be picklable.

            FIXME: This function is incomplete! Not all necessary members are serialized
        """
        if not self.is_normalized:
            raise ValueError("You cannot serialize queries unless they are normalized")

        result = {}
        result["kind"] = self.kind
        result["table"] = self.model._meta.db_table
        result["concrete_table"] = self.concrete_model._meta.db_table
        result["columns"] = list(self.columns or [])  # set() is not JSONifiable
        result["projection_possible"] = self.projection_possible
        result["init_list"] = self.init_list
        result["distinct"] = self.distinct
        result["order_by"] = self.order_by
        result["low_mark"] = self.low_mark
        result["high_mark"] = self.high_mark
        result["excluded_pks"] = list(map(str, self.excluded_pks))

        where = []

        if self.where:
            assert self.where.connector == "OR"
            for node in self.where.children:
                assert node.connector == "AND"

                query = {}

                if node.children:
                    for lookup in node.children:

                        query["".join([lookup.column, lookup.operator])] = _serialize_sql_value(lookup.value)
                else:
                    query["".join([node.column, node.operator])] = _serialize_sql_value(node.value)

                where.append(query)

        result["where"] = where

        return json.dumps(result)


INVALID_ORDERING_FIELD_MESSAGE = (
    "Ordering on TextField or BinaryField is not supported on the rpc. "
    "You might consider using a ComputedCharField which stores the first "
    "_MAX_STRING_LENGTH (from google.appengine.api.datastore_types) bytes of the "
    "field and instead order on that."
)


def _serialize_sql_value(value):
    if isinstance(value, int):
        return value
    else:
        return str("NULL" if value is None else value)


def _get_parser(query, connection=None):
    from gcloudc.db.backends.datastore.parsers import base

    return base.BaseParser(query, connection)


def transform_query(connection, query):
    return _get_parser(query, connection).get_transformed_query()


def extract_ordering(query):
    return _get_parser(query).get_extracted_ordering()



from django.db.models.expressions import Star
from django.db.models.sql.datastructures import EmptyResultSet
from django.db.models.sql.where import SubqueryConstraint

from .version_19 import Parser as BaseParser


class Parser(BaseParser):
    def _prepare_for_transformation(self):
        from django.db.models.sql.where import EmptyWhere

        if isinstance(self.django_query.where, EmptyWhere):
            # Empty where means return nothing!
            raise EmptyResultSet()

    def _where_node_leaf_callback(self, node, negated, new_parent, connection, model, compiler):
        if not isinstance(node, SubqueryConstraint):
            # Only do this test if it's not a subquery, the parent method deals with that
            if not hasattr(node, "lhs") and not hasattr(node, "rhs"):
                # Empty Q() object - basically an empty where node that needs nothing doing to it
                return

        return super(Parser, self)._where_node_leaf_callback(node, negated, new_parent, connection, model, compiler)

    def _determine_query_kind(self):
        query = self.django_query
        if query.annotations:
            if "__count" in query.annotations:
                field = query.annotations["__count"].input_field
                if isinstance(field, Star) or field.value == "*":
                    return "COUNT"

        return "SELECT"





from django.db import connection as default_connection
from django.db import connections
from django.db.models.query import Q
from django.core.exceptions import EmptyResultSet

from gcloudc.db.backends.datastore import transaction
from gcloudc.db.backends.datastore.dnf import normalize_query
from gcloudc.db.backends.datastore.query import (
    Query,
    WhereNode,
    transform_query,
)

from . import TestCase
from .models import (
    InheritedModel,
    Relation,
    TestUser,
    TransformTestModel,
)

DEFAULT_NAMESPACE = default_connection.ops.connection.settings_dict.get("NAMESPACE")


def find_children_containing_node(list_of_possible_children, column, op, value):
    for children in list_of_possible_children:
        for node in (children or []):
            if node.column == column and node.operator == op and node.value == value:
                return children


class TransformQueryTest(TestCase):

    def test_polymodel_filter_applied(self):
        query = transform_query(
            connections['default'],
            InheritedModel.objects.filter(field1="One").all().query
        )
        query.prepare()

        self.assertEqual(2, len(query.where.children))
        self.assertTrue(query.where.children[0].children[0].is_leaf)
        self.assertTrue(query.where.children[1].children[0].is_leaf)
        self.assertEqual("class", query.where.children[0].children[0].column)
        self.assertEqual("field1", query.where.children[1].children[0].column)

    def test_basic_query(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.all().query
        )

        self.assertEqual(query.model, TransformTestModel)
        self.assertEqual(query.kind, 'SELECT')
        self.assertEqual(query.tables, [TransformTestModel._meta.db_table])
        self.assertIsNone(query.where)

    def test_and_filter(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.filter(field1="One", field2="Two").all().query
        )

        self.assertEqual(query.model, TransformTestModel)
        self.assertEqual(query.kind, 'SELECT')
        self.assertEqual(query.tables, [TransformTestModel._meta.db_table])
        self.assertTrue(query.where)
        self.assertEqual(2, len(query.where.children))  # Two child nodes

    def test_exclude_filter(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.exclude(field1="One").all().query
        )

        self.assertEqual(query.model, TransformTestModel)
        self.assertEqual(query.kind, 'SELECT')
        self.assertEqual(query.tables, [TransformTestModel._meta.db_table])
        self.assertTrue(query.where)
        self.assertEqual(1, len(query.where.children))  # One child node
        self.assertTrue(query.where.children[0].negated)
        self.assertEqual(1, len(query.where.children[0].children))

    def test_ordering(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.filter(field1="One", field2="Two").order_by("field1", "-field2").query
        )

        self.assertEqual(query.model, TransformTestModel)
        self.assertEqual(query.kind, 'SELECT')
        self.assertEqual(query.tables, [TransformTestModel._meta.db_table])
        self.assertTrue(query.where)
        self.assertEqual(2, len(query.where.children))  # Two child nodes
        self.assertEqual(["field1", "-field2"], query.order_by)

    def test_projection(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.only("field1").query
        )

        self.assertItemsEqual(["id", "field1"], query.columns)

        query = transform_query(
            connections['default'],
            TransformTestModel.objects.values_list("field1").query
        )

        self.assertEqual(set(["field1"]), query.columns)

        query = transform_query(
            connections['default'],
            TransformTestModel.objects.defer("field1", "field4").query
        )

        self.assertItemsEqual(set(["id", "field2", "field3"]), query.columns)

    def test_no_results_returns_emptyresultset(self):
        self.assertRaises(
            EmptyResultSet,
            transform_query,
            connections['default'],
            TransformTestModel.objects.none().query
        )

    def test_offset_and_limit(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.all()[5:10].query
        )

        self.assertEqual(5, query.low_mark)
        self.assertEqual(10, query.high_mark)

    def test_isnull(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.filter(field3__isnull=True).all()[5:10].query
        )

        self.assertTrue(query.where.children[0].value)
        self.assertEqual("ISNULL", query.where.children[0].operator)

    def test_distinct(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.distinct("field2", "field3").query
        )

        self.assertTrue(query.distinct)
        self.assertEqual(query.columns, set(["field2", "field3"]))

        query = transform_query(
            connections['default'],
            TransformTestModel.objects.distinct().values("field2", "field3").query
        )

        self.assertTrue(query.distinct)
        self.assertEqual(query.columns, set(["field2", "field3"]))

    def test_order_by_pk(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.order_by("pk").query
        )

        self.assertEqual("__key__", query.order_by[0])

        query = transform_query(
            connections['default'],
            TransformTestModel.objects.order_by("-pk").query
        )

        self.assertEqual("-__key__", query.order_by[0])

    def test_reversed_ordering(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.order_by("pk").reverse().query
        )

        self.assertEqual("-__key__", query.order_by[0])

    def test_clear_ordering(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.order_by("pk").order_by().query
        )

        self.assertFalse(query.order_by)

    def test_projection_on_textfield_disabled(self):
        query = transform_query(
            connections['default'],
            TransformTestModel.objects.values_list("field4").query
        )

        self.assertFalse(query.columns)
        self.assertFalse(query.projection_possible)


class QueryNormalizationTests(TestCase):
    """
        The parse_dnf function takes a Django where tree, and converts it
        into a tree of one of the following forms:

        [ (column, operator, value), (column, operator, value) ] <- AND only query
        [ [(column, operator, value)], [(column, operator, value) ]] <- OR query, of multiple ANDs
    """

    def test_and_with_child_or_promoted(self):
        """
            Given the following tree:

                   AND
                  / | ﹨
                 A  B OR
                      / ﹨
                     C   D

             The OR should be promoted, so the resulting tree is

                      OR
                     /   ﹨
                   AND   AND
                  / | ﹨ / | ﹨
                 A  B C A B D
        """

        query = Query(TestUser, "SELECT")
        query.where = WhereNode('default')
        query.where.children.append(WhereNode('default'))
        query.where.children[-1].column = "A"
        query.where.children[-1].operator = "="
        query.where.children.append(WhereNode('default'))
        query.where.children[-1].column = "B"
        query.where.children[-1].operator = "="
        query.where.children.append(WhereNode('default'))
        query.where.children[-1].connector = "OR"
        query.where.children[-1].children.append(WhereNode('default'))
        query.where.children[-1].children[-1].column = "C"
        query.where.children[-1].children[-1].operator = "="
        query.where.children[-1].children.append(WhereNode('default'))
        query.where.children[-1].children[-1].column = "D"
        query.where.children[-1].children[-1].operator = "="

        query = normalize_query(query)

        self.assertEqual(query.where.connector, "OR")
        self.assertEqual(2, len(query.where.children))
        self.assertFalse(query.where.children[0].is_leaf)
        self.assertFalse(query.where.children[1].is_leaf)
        self.assertEqual(query.where.children[0].connector, "AND")
        self.assertEqual(query.where.children[1].connector, "AND")
        self.assertEqual(3, len(query.where.children[0].children))
        self.assertEqual(3, len(query.where.children[1].children))

    def test_and_queries(self):
        qs = TestUser.objects.filter(username="test").all()

        query = normalize_query(transform_query(
            connections['default'],
            qs.query
        ))

        self.assertTrue(1, len(query.where.children))
        self.assertEqual(query.where.children[0].children[0].column, "username")
        self.assertEqual(query.where.children[0].children[0].operator, "=")
        self.assertEqual(query.where.children[0].children[0].value, "test")

        qs = TestUser.objects.filter(username="test", email="test@example.com")

        query = normalize_query(transform_query(
            connections['default'],
            qs.query
        ))

        self.assertTrue(2, len(query.where.children[0].children))

        self.assertEqual(query.where.connector, "OR")
        self.assertEqual(query.where.children[0].connector, "AND")

        self.assertTrue(find_children_containing_node([query.where.children[0]], "username", "=", "test"))
        self.assertTrue(find_children_containing_node([query.where.children[0]], "email", "=", "test@example.com"))

        qs = TestUser.objects.filter(username="test").exclude(email="test@example.com")
        query = normalize_query(transform_query(
            connections['default'],
            qs.query
        ))

        self.assertTrue(2, len(query.where.children[0].children))
        self.assertEqual(query.where.connector, "OR")
        self.assertEqual(query.where.children[0].connector, "AND")

        possible_children = (query.where.children[0], query.where.children[1])

        expected = find_children_containing_node(possible_children, "email", "<", "test@example.com")
        self.assertTrue(find_children_containing_node([expected], "username", "=", "test"))
        self.assertTrue(find_children_containing_node([expected], "email", "<", "test@example.com"))

        expected = find_children_containing_node(possible_children, "email", ">", "test@example.com")
        self.assertTrue(find_children_containing_node([expected], "username", "=", "test"))
        self.assertTrue(find_children_containing_node([expected], "email", ">", "test@example.com"))

        instance = Relation(pk=1)
        qs = instance.related_set.filter(headline__startswith='Fir')

        query = normalize_query(transform_query(
            connections['default'],
            qs.query
        ))

        self.assertTrue(2, len(query.where.children[0].children))
        self.assertEqual(query.where.connector, "OR")
        self.assertEqual(query.where.children[0].connector, "AND")
        self.assertEqual(query.where.children[0].children[0].column, "relation_id")
        self.assertEqual(query.where.children[0].children[0].operator, "=")
        self.assertEqual(query.where.children[0].children[0].value, 1)
        self.assertEqual(query.where.children[0].children[1].column, "_idx_startswith_headline")
        self.assertEqual(query.where.children[0].children[1].operator, "=")
        self.assertEqual(query.where.children[0].children[1].value, u"Fir")

    def test_impossible_or_query(self):
        qs = TestUser.objects.filter(
            username="python").filter(
            Q(username__in=["ruby", "jruby"]) | (Q(username="php") & ~Q(username="perl"))
        )

        with self.assertRaises(EmptyResultSet):
            normalize_query(transform_query(
                connections['default'],
                qs.query
            ))

    def test_or_queries(self):
        qs = TestUser.objects.filter(
            first_name="python").filter(
            Q(username__in=["ruby", "jruby"]) | (Q(username="php") & ~Q(username="perl"))
        )

        query = normalize_query(transform_query(
            connections['default'],
            qs.query
        ))

        # After IN and != explosion, we have...
        # (AND:
        #       (first_name='python',
        #        OR: (username='ruby', username='jruby',
        #             AND: (username='php',
        #                   AND: (username < 'perl', username > 'perl')
        #                  )
        #            )
        #        )
        # )

        # Working backwards,
        # AND: (username < 'perl', username > 'perl') can't be simplified
        #
        # AND: (username='php', AND: (username < 'perl', username > 'perl'))
        # can become
        # (OR: (AND: username = 'php', username < 'perl'), (AND: username='php', username > 'perl'))
        #
        # OR: (username='ruby', username='jruby',(OR: (AND: username = 'php', username < 'perl'),
        # (AND: username='php', username > 'perl')) can't be simplified
        #
        # (AND: (first_name='python', OR: (username='ruby', username='jruby',
        # (OR: (AND: username = 'php', username < 'perl'), (AND: username='php', username > 'perl'))
        # becomes...
        # (OR: (AND: first_name='python', username = 'ruby'), (AND: first_name='python', username='jruby'),
        #      (AND: first_name='python', username='php', username < 'perl')
        #      (AND: first_name='python', username='php', username > 'perl')

        self.assertTrue(4, len(query.where.children[0].children))

        self.assertEqual(query.where.children[0].connector, "AND")

        possible_children = [query.where.children[i] for i in range(4)]

        expected = find_children_containing_node(possible_children, "username", "=", "jruby")
        self.assertTrue(find_children_containing_node([expected], "first_name", "=", "python"))
        self.assertTrue(find_children_containing_node([expected], "username", "=", "jruby"))

        expected = find_children_containing_node(possible_children, "username", ">", "perl")
        self.assertTrue(find_children_containing_node([expected], "first_name", "=", "python"))
        self.assertTrue(find_children_containing_node([expected], "username", "=", "php"))
        self.assertTrue(find_children_containing_node([expected], "username", ">", "perl"))

        expected = find_children_containing_node(possible_children, "username", "<", "perl")
        self.assertTrue(find_children_containing_node([expected], "first_name", "=", "python"))
        self.assertTrue(find_children_containing_node([expected], "username", "=", "php"))
        self.assertTrue(find_children_containing_node([expected], "username", "<", "perl"))

        expected = find_children_containing_node(possible_children, "username", "=", "ruby")
        self.assertTrue(find_children_containing_node([expected], "first_name", "=", "python"))
        self.assertTrue(find_children_containing_node([expected], "username", "=", "ruby"))

        qs = TestUser.objects.filter(username="test") | TestUser.objects.filter(username="cheese")

        query = normalize_query(transform_query(
            connections['default'],
            qs.query
        ))

        self.assertEqual(query.where.connector, "OR")
        self.assertEqual(2, len(query.where.children))
        self.assertTrue(query.where.children[0].is_leaf)
        self.assertTrue("cheese" in (query.where.children[0].value, query.where.children[1].value))
        self.assertTrue(query.where.children[1].is_leaf)
        self.assertTrue("test" in (query.where.children[0].value, query.where.children[1].value))

        qs = TestUser.objects.using("default").filter(username__in=set()).values_list('email')

        with self.assertRaises(EmptyResultSet):
            query = normalize_query(transform_query(
                connections['default'],
                qs.query
            ))

        qs = TestUser.objects.filter(
            username__startswith='Hello'
        ) | TestUser.objects.filter(username__startswith='Goodbye')

        query = normalize_query(transform_query(
            connections['default'],
            qs.query
        ))

        self.assertEqual(2, len(query.where.children))
        self.assertEqual("_idx_startswith_username", query.where.children[0].column)
        self.assertTrue("Goodbye" in (query.where.children[0].value, query.where.children[1].value))
        self.assertEqual("_idx_startswith_username", query.where.children[1].column)
        self.assertTrue("Hello" in (query.where.children[0].value, query.where.children[1].value))

        qs = TestUser.objects.filter(pk__in=[1, 2, 3])
        query = normalize_query(transform_query(
            connections['default'],
            qs.query
        ))

        rpc = transaction._rpc(default_connection.alias)

        self.assertEqual(3, len(query.where.children))
        self.assertEqual("__key__", query.where.children[0].column)
        self.assertEqual("__key__", query.where.children[1].column)
        self.assertEqual("__key__", query.where.children[2].column)
        self.assertEqual({
                rpc.key(TestUser._meta.db_table, 1),
                rpc.key(TestUser._meta.db_table, 2),
                rpc.key(TestUser._meta.db_table, 3),
            }, {
                query.where.children[0].value,
                query.where.children[1].value,
                query.where.children[2].value,
            }
        )

        qs = TestUser.objects.filter(pk__in=[1, 2, 3]).filter(username="test")
        query = normalize_query(transform_query(
            connections['default'],
            qs.query
        ))

        self.assertEqual(3, len(query.where.children))
        self.assertEqual("__key__", query.where.children[0].children[0].column)
        self.assertEqual("test", query.where.children[0].children[1].value)

        self.assertEqual("__key__", query.where.children[1].children[0].column)
        self.assertEqual("test", query.where.children[0].children[1].value)

        self.assertEqual("__key__", query.where.children[2].children[0].column)
        self.assertEqual("test", query.where.children[0].children[1].value)

        self.assertEqual({
                rpc.key(TestUser._meta.db_table, 1),
                rpc.key(TestUser._meta.db_table, 2),
                rpc.key(TestUser._meta.db_table, 3),
            }, {
                query.where.children[0].children[0].value,
                query.where.children[1].children[0].value,
                query.where.children[2].children[0].value,
            }
        )

    def test_removal_of_multiple_pk_equalities(self):
        """ Regression test for #1174/#1175.
            Make sure that we don't get an error when a query has multiple different equality
            filters on the PK.
        """
        query = TransformTestModel.objects.filter(pk=1).filter(pk=2).filter(pk=3)
        try:
            list(query)
        except ValueError:
            raise
            self.fail("ValueError raised when filtering on multiple different PK equalities")



# LIBRARIES
from django.db.models.expressions import (
    OrderBy,
    Value,
)
from django.db.models.sql import compiler
from django.db.models.sql.query import get_order_dir

# GCLOUDC
from .commands import (
    DeleteCommand,
    InsertCommand,
    SelectCommand,
    UpdateCommand,
)


class SQLCompiler(compiler.SQLCompiler):
    def find_ordering_name(self, name, opts, alias=None, default_order="ASC", already_seen=None):
        """
            Overridden just for the __scatter__ property ordering
        """

        # This allow special appengine properties (e.g. __scatter__) to be supplied as an ordering
        # even though they don't (and can't) exist as Django model fields
        if name.startswith("__") and name.endswith("__"):
            name, order = get_order_dir(name, default_order)
            descending = True if order == "DESC" else False
            return [(OrderBy(Value("__scatter__"), descending=descending), False)]

        return super(SQLCompiler, self).find_ordering_name(
            name, opts, alias=alias, default_order=default_order, already_seen=already_seen
        )

    def as_sql(self, with_limits=True, with_col_aliases=False, subquery=False):
        self.pre_sql_setup()
        self.refcounts_before = self.query.alias_refcount.copy()

        select = SelectCommand(self.connection, self.query)
        return (select, tuple())

    def get_select(self):
        self.query.select_related = False  # Make sure select_related is disabled for all queries
        return super(SQLCompiler, self).get_select()


class SQLInsertCompiler(SQLCompiler, compiler.SQLInsertCompiler):
    def __init__(self, *args, **kwargs):
        self.return_id = None
        super(SQLInsertCompiler, self).__init__(*args, **kwargs)

    def as_sql(self, with_limits=True, with_col_aliases=False, subquery=False):
        self.pre_sql_setup()

        from gcloudc.db.backends.datastore.utils import get_concrete_fields

        # Always pass down all the fields on an insert
        return [
            (
                InsertCommand(
                    self.connection,
                    self.query.model,
                    self.query.objs,
                    list(self.query.fields) + list(get_concrete_fields(self.query.model, ignore_leaf=True)),
                    self.query.raw,
                ),
                tuple(),
            )
        ]


class SQLDeleteCompiler(SQLCompiler, compiler.SQLDeleteCompiler):
    def as_sql(self, with_limits=True, with_col_aliases=False, subquery=False):
        return (DeleteCommand(self.connection, self.query), tuple())


class SQLUpdateCompiler(SQLCompiler, compiler.SQLUpdateCompiler):
    def as_sql(self, with_limits=True, with_col_aliases=False, subquery=False):
        self.pre_sql_setup()
        return (UpdateCommand(self.connection, self.query), tuple())


class SQLAggregateCompiler(compiler.SQLAggregateCompiler, SQLCompiler):
    def as_sql(self, with_limits=True, with_col_aliases=False, subquery=False):
        if self.query.subquery:
            self.query.high_mark = self.query.subquery.query.high_mark
            self.query.low_mark = self.query.subquery.query.low_mark
        return SQLCompiler.as_sql(self, with_limits, with_col_aliases, subquery)



import copy
import decimal
import logging
from datetime import datetime

import django
from django.db import (
    DatabaseError,
    IntegrityError,
)

from django.utils.encoding import (
    force_str,
)
from google.cloud.datastore.entity import Entity
from google.cloud.datastore.key import Key
from google.cloud.datastore.query import Query

from . import (
    POLYMODEL_CLASS_ATTRIBUTE,
    caching,
    meta_queries,
    transaction,
    utils,
)
from .caching import remove_entities_from_cache_by_key
from .constraints import (
    CONSTRAINT_VIOLATION_MSG,
    check_unique_markers_in_memory,
    has_active_unique_constraints,
)
from .dbapi import NotSupportedError
from .dnf import normalize_query
from .formatting import generate_sql_representation
from .query import transform_query
from .query_utils import (
    get_filter,
    has_filter,
)
from .unique_utils import (
    _unique_combinations,
    query_is_unique,
)
from .utils import (
    MockInstance,
    django_instance_to_entities,
    ensure_datetime,
    get_datastore_key,
    get_field_from_column,
    has_concrete_parents,
)

logger = logging.getLogger(__name__)

OPERATORS_MAP = {
    "exact": "=",
    "gt": ">",
    "gte": ">=",
    "lt": "<",
    "lte": "<=",
    # The following operators are supported with special code below.
    "isnull": None,
    "in": None,
    "range": None,
}

EXTRA_SELECT_FUNCTIONS = {
    "+": lambda x, y: x + y,
    "-": lambda x, y: x - y,
    "/": lambda x, y: x / y,
    "*": lambda x, y: x * y,
    "<": lambda x, y: x < y,
    ">": lambda x, y: x > y,
    "=": lambda x, y: x == y,
}

REVERSE_OP_MAP = {"=": "exact", ">": "gt", ">=": "gte", "<": "lt", "<=": "lte"}

INEQUALITY_OPERATORS = frozenset([">", "<", "<=", ">="])


def _cols_from_where_node(where_node):
    cols = where_node.get_cols() if hasattr(where_node, "get_cols") else where_node.get_group_by_cols()
    return cols


def _get_tables_from_where(where_node):
    cols = _cols_from_where_node(where_node)
    if django.VERSION[1] < 8:
        return list(set([x[0] for x in cols if x[0]]))
    else:
        return list(set([x.alias for x in cols]))


def field_conv_year_only(value):
    value = ensure_datetime(value)
    return datetime(value.year, 1, 1, 0, 0)


def field_conv_month_only(value):
    value = ensure_datetime(value)
    return datetime(value.year, value.month, 1, 0, 0)


def field_conv_day_only(value):
    value = ensure_datetime(value)
    return datetime(value.year, value.month, value.day, 0, 0)


def coerce_unicode(value):
    if isinstance(value, bytes):
        try:
            value = value.decode("utf-8")
        except UnicodeDecodeError:
            # This must be a Django databaseerror, because the exception happens too
            # early before Django's exception wrapping can take effect (e.g. it happens on SQL
            # construction, not on execution.
            raise DatabaseError("Bytestring is not encoded in utf-8")

    # The SDK raises BadValueError for unicode sub-classes like SafeText.
    return str(value)


def log_once(logging_call, text, args):
    """
        Only logs one instance of the combination of text and arguments to the passed
        logging function
    """
    identifier = "%s:%s" % (text, args)
    if identifier in log_once.logged:
        return
    logging_call(text % args)
    log_once.logged.add(identifier)


log_once.logged = set()


def convert_django_ordering_to_gae(ordering):
    return ordering


def limit_results_generator(results, limit):
    for result in results:
        yield result
        limit -= 1
        if not limit:
            raise StopIteration


def can_perform_datastore_get(normalized_query):
    """
        Given a normalized query, returns True if there is an equality
        filter on a key in each branch of the where
    """
    assert normalized_query.is_normalized

    for and_branch in normalized_query.where.children:
        if and_branch.is_leaf:
            if and_branch.column != "__key__" or and_branch.operator != "=":
                return False
        else:
            key_found = False
            for filter_node in and_branch.children:
                assert filter_node.is_leaf

                if filter_node.column == "__key__":
                    if filter_node.operator == "=":
                        key_found = True
                        break

            if not key_found:
                return False

    return True


class EntityTransforms:
    @staticmethod
    def convert_key_to_entity(result):
        class FakeEntity(dict):
            def __init__(self, key):
                self._key = key

            @property
            def key(self):
                return self._key

        return FakeEntity(result)

    @staticmethod
    def rename_pk_field(model, concrete_model, result):
        if result is None:
            return result

        value = result.key.id_or_name
        result[model._meta.pk.column] = value
        result[concrete_model._meta.pk.column] = value
        return result

    @staticmethod
    def process_extra_selects(query, result):
        """
            We handle extra selects by generating the new columns from
            each result. We can handle simple boolean logic and operators.
        """
        if result is None:
            return result

        extra_selects = query.extra_selects
        model_fields = query.model._meta.fields

        DATE_FORMATS = ("%Y-%m-%d", "%Y-%m-%d %H:%M:%S")

        def process_arg(arg):
            if arg.startswith("'") and arg.endswith("'"):
                # String literal
                arg = arg.strip("'")
                # Check to see if this is a date
                for date in DATE_FORMATS:
                    try:
                        value = datetime.strptime(arg, date)
                        return value
                    except ValueError:
                        continue
                return arg
            elif arg in [x.column for x in model_fields]:
                # Column value
                return result.get(arg)

            # Handle NULL
            if arg.lower() == "null":
                return None
            elif arg.lower() == "true":
                return True
            elif arg.lower() == "false":
                return False

            # See if it's an integer
            try:
                arg = int(arg)
            except (TypeError, ValueError):
                pass

            # Just a plain old literal
            return arg

        for col, select in extra_selects:
            result[col] = select[0](*[process_arg(x) for x in select[1]])

        return result

    @staticmethod
    def convert_datetime_fields(query, result):
        if result is None:
            return result

        fields = [
            x for x in query.model._meta.fields if x.get_internal_type() in ("DateTimeField", "DateField", "TimeField")
        ]

        for field in fields:
            column = field.column
            if isinstance(result, dict):  # sometimes it's a key!
                value = result.get(column)
            else:
                value = None

            if value is not None:
                result[column] = ensure_datetime(value)

        return result

    @staticmethod
    def fix_projected_values_type(query, result):
        """
            String values returned from projection queries return as 'str' not 'unicode'
            See https://github.com/potatolondon/djangae/issues/1026

            FIXME: This was the case on App Engine, probably not on Cloud Datastore. When
            all original Djangae tests pass, let's remove this function and see if they still pass!
        """

        if result is None:
            return None

        fields = [x for x in query.model._meta.fields if x.get_internal_type() in ("CharField",)]

        for field in fields:
            col = field.column
            if col in result and isinstance(result[col], bytes):
                result[col] = str(result[col], "utf-8")

        return result

    @staticmethod
    def ignore_excluded_pks(excluded_pks, result):
        if result is None:
            return result

        if result.key in excluded_pks:
            return None

        return result


class SelectCommand(object):
    def __init__(self, connection, query, keys_only=False):
        self.connection = connection.alias
        self.namespace = connection.namespace

        self.query = transform_query(connection, query)
        self.query.prepare()
        self.query = normalize_query(self.query)

        self.original_query = query

        # We enable keys only queries if they have been forced, or, if
        # someone did only("pk") or someone did values_list("pk") this is a little
        # inconsistent with other fields which aren't projected if just values(_list) is used
        self.keys_only = (
            keys_only
            or (
                query.deferred_loading[1] is False
                and len(query.deferred_loading[0]) == 1
                and query.model._meta.pk.column in query.deferred_loading[0]
            )
            or (len(query.select) == 1 and query.select[0].field == query.model._meta.pk)
        )

        # MultiQuery doesn't support keys_only
        if self.query.where and len(self.query.where.children) > 1:
            self.keys_only = False

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.query.serialize() == other.query.serialize()

    def __ne__(self, other):
        return not self.__eq__(other)

    def _sanity_check(self):
        if self.query.distinct and not self.query.columns:
            raise NotSupportedError("Tried to perform distinct query when projection wasn't possible")

    def _exclude_pk(self, columns):
        if columns is None:
            return None

        opts = self.query.model._meta
        copts = self.query.concrete_model._meta

        return [x for x in columns if x not in (opts.pk.column, copts.pk.column)]

    def _build_query(self):
        self._sanity_check()

        queries = []
        projection = self._exclude_pk(self.query.columns) or None

        # FIXME: Should this always be the projection fields if distinct is True?
        query_kwargs = {
            "kind": self.query.concrete_model._meta.db_table,
            "distinct_on": projection if self.query.distinct else (),
            "projection": projection or (),
            "namespace": self.namespace,
        }

        ordering = convert_django_ordering_to_gae(self.query.order_by)

        if self.query.distinct and not ordering:
            # If we specified we wanted a distinct query, but we didn't specify
            # an ordering, we must set the ordering to the distinct columns, otherwise
            # App Engine shouts at us. Nastily. And without remorse.
            # The order of the columns in `ordering` makes a difference, but `distinct` is a set
            # and therefore unordered, but in  this situation (where the ordering has not been
            # explicitly defined) any order of the columns will do
            ordering = list(self.query.columns)

        # Deal with the no filters case
        if self.query.where is None:
            query = transaction._rpc(self.connection).query(**query_kwargs)

            if self.keys_only:
                query.keys_only()

            query.order = ordering
            return query

        assert self.query.where

        rpc = transaction._rpc(self.connection)
        # Go through the normalized query tree
        for and_branch in self.query.where.children:
            query = rpc.query(**query_kwargs)

            if self.keys_only:
                query.keys_only()

            # This deals with the oddity that the root of the tree may well be a leaf
            filters = [and_branch] if and_branch.is_leaf else and_branch.children

            for filter_node in filters:
                lookup = (filter_node.column, filter_node.operator)

                value = filter_node.value

                # This is a special case. Annoyingly Django's decimal field doesn't
                # ever call ops.get_prep_save or lookup or whatever when you are filtering
                # on a query. It *does* do it on a save, so we basically need to do a
                # conversion here, when really it should be handled elsewhere
                if isinstance(value, decimal.Decimal):
                    field = get_field_from_column(self.query.model, filter_node.column)
                    value = self.connection.ops.adapt_decimalfield_value(value, field.max_digits, field.decimal_places)
                elif isinstance(value, (str, bytes)):
                    value = coerce_unicode(value)
                elif isinstance(value, Key):
                    # Make sure we apply the current namespace to any lookups
                    # by key. Fixme: if we ever add key properties this will break if
                    # someone is trying to filter on a key which has a different namespace
                    # to the active one.
                    value = rpc.key(value.kind, value.id_or_name)

                # If there is already a value for this lookup, we need to make the
                # value a list and append the new entry
                filter_value = get_filter(query, lookup)
                if has_filter(query, lookup) and not isinstance(filter_value, (list, tuple)) and filter_value != value:
                    new_value = [filter_value] + [value]
                    query.add_filter(lookup[0], lookup[1], new_value)
                else:
                    # If the value is a list, we can't just assign it to the query
                    # which will treat each element as its own value. So in this
                    # case we nest it. This has the side effect of throwing a BadValueError
                    # which we could throw ourselves, but the datastore might start supporting
                    # list values in lookups.. you never know!
                    # FIXME: I can't remember the reason for this rather than actually just
                    # throwing an error?
                    if isinstance(value, (list, tuple)):
                        query.add_filter(lookup[0], lookup[1], [value])
                    else:
                        # Common case: just add the raw where constraint
                        query.add_filter(lookup[0], lookup[1], value)

            if ordering:
                query.order = ordering

            queries.append(query)

        if can_perform_datastore_get(self.query):
            # Yay for optimizations!
            return meta_queries.QueryByKeys(self.connection, self.query.model, queries, ordering, self.namespace)

        if len(queries) == 1:
            identifier = query_is_unique(self.query.model, queries[0])
            if identifier:
                # Yay for optimizations!
                return meta_queries.UniqueQuery(identifier, queries[0], self.query.model, self.namespace)

            return queries[0]
        else:
            return meta_queries.AsyncMultiQuery(queries, ordering)

    def _fetch_results(self, query):
        # If we're manually excluding PKs, and we've specified a limit to the results
        # we need to make sure that we grab more than we were asked for otherwise we could filter
        # out too many! These are again limited back to the original request limit
        # while we're processing the results later
        # Apply the namespace before excluding
        rpc = transaction._rpc(self.connection)

        excluded_pks = [rpc.key(x.kind, x.id_or_name) for x in self.query.excluded_pks]

        high_mark = self.query.high_mark
        low_mark = self.query.low_mark

        excluded_pk_count = 0
        if excluded_pks and high_mark:
            excluded_pk_count = len(excluded_pks)
            high_mark += excluded_pk_count

        limit = None if high_mark is None else (high_mark - (low_mark or 0))
        offset = low_mark or 0

        if self.query.kind == "COUNT":
            if excluded_pks:
                # If we're excluding pks, relying on a traditional count won't work
                # so we have two options:
                # 1. Do a keys_only query instead and count the results excluding keys
                # 2. Do a count, then a pk__in=excluded_pks to work out how many to subtract
                # Here I've favoured option one as it means a single RPC call. Testing locally
                # didn't seem to indicate much of a performance difference, even when doing the pk__in
                # with GetAsync while the count was running. That might not be true of prod though so
                # if anyone comes up with a faster idea let me know!
                if isinstance(query, meta_queries.QueryByKeys):
                    # If this is a QueryByKeys, just do the datastore Get and count the results
                    resultset = (x.key for x in query.fetch(limit=limit, offset=offset) if x)
                else:
                    count_query = Query(query._Query__kind, keys_only=True, namespace=self.namespace)
                    count_query.update(query)
                    resultset = count_query.Run(limit=limit, offset=offset)
                self.results = [len([y for y in resultset if y not in excluded_pks])]
                self.results_returned = 1
            else:
                query.keys_only()

                self.results = [len(list(query.fetch(limit=limit, offset=offset)))]
                self.results_returned = 1
            return
        elif self.query.kind == "AVERAGE":
            raise ValueError("AVERAGE not yet supported")

        # Ensure that the results returned is reset
        self.results_returned = 0
        self.results = []

        seen = set()

        def dedupe(result):
            # FIXME: This logic can't be right. I think we need to store the distinct fields
            # somewhere on the query
            if getattr(self.original_query, "annotation_select", None):
                columns = self.original_query.annotation_select.keys()
            else:
                columns = self.query.columns or []
            if not columns:
                return result

            key = tuple([result[x] for x in self._exclude_pk(columns) if x in result])
            if key in seen:
                return None
            seen.add(key)
            return result

        for entity in query.fetch(limit=limit, offset=offset):
            # If this is a keys only query, we need to generate a fake entity
            # for each key in the result set
            if isinstance(entity, Key):
                entity = EntityTransforms.convert_key_to_entity(entity)

            entity = EntityTransforms.ignore_excluded_pks(excluded_pks, entity)
            entity = EntityTransforms.convert_datetime_fields(self.query, entity)
            entity = EntityTransforms.fix_projected_values_type(self.query, entity)
            entity = EntityTransforms.rename_pk_field(self.query.model, self.query.concrete_model, entity)
            entity = EntityTransforms.process_extra_selects(self.query, entity)

            if self.query.distinct and self.query.extra_selects:
                entity = dedupe(entity)

            if entity:
                self.results.append(entity)
                self.results_returned += 1

            if limit and self.results_returned >= (limit - excluded_pk_count):
                break

    def execute(self):
        self.gae_query = self._build_query()
        self._fetch_results(self.gae_query)
        self.results = iter(self.results)
        return self.results_returned

    def __repr__(self):
        return force_str(generate_sql_representation(self))

    def __mod__(self, params):
        return repr(self)

    def lower(self):
        """
            This exists solely for django-debug-toolbar compatibility.
        """
        return str(self).lower()


class FlushCommand(object):
    """
        sql_flush returns the SQL statements to flush the database,
        which are then executed by cursor.execute()

        We instead return a list of FlushCommands which are called by
        our cursor.execute
    """

    def __init__(self, table, connection):
        self.connection = connection.alias
        self.table = table
        self.namespace = connection.namespace

    def execute(self):
        table = self.table
        query = transaction._rpc(self.connection).query(kind=table, namespace=self.namespace)
        query.keys_only()

        # The local datastore emulator explodes if you try to delete more than 500
        # things in a single batch. So we just iterate in batches of 500.
        limit = 500

        results = [x.key for x in query.fetch(limit=limit)]
        while results:
            transaction._rpc(self.connection).delete(results)
            results = [x.key for x in query.fetch(limit=limit)]


def reserve_id(connection, kind, id_or_name, namespace):
    if not isinstance(id_or_name, int):
        # Nothing to do if the ID is a string, no-need to reserve that
        return

    gclient = connection.connection.gclient
    gclient.reserve_ids_sequential(gclient.key(kind, id_or_name, namespace=namespace), 1)


def perform_unique_checks(namespace, model, rpc, primary, test_fn):
    combinations = _unique_combinations(model, ignore_pk=True)
    for combination in combinations:
        query = rpc.query(kind=primary.kind, namespace=namespace)

        for field in combination:
            col_name = model._meta.get_field(field).column
            value = primary.get(col_name)
            if isinstance(value, list):
                for item in value:
                    query.add_filter(col_name, '=', item)
            elif value is not None:
                query.add_filter(col_name, '=', value)

        # only perform the query if there are filters on it
        if len(query.filters):
            res = list(query.fetch(1))
            if test_fn(res):
                raise IntegrityError(CONSTRAINT_VIOLATION_MSG.format(model._meta.db_table, ", ".join(combination)))


class BulkInsertError(IntegrityError, NotSupportedError):
    pass


class BulkDeleteError(IntegrityError, NotSupportedError):
    pass


class InsertCommand(object):
    def __init__(self, connection, model, objs, fields, raw):
        self.has_pk = any(x.primary_key for x in fields)
        self.model = model
        self.objs = objs
        self.connection = connection
        self.namespace = connection.namespace
        self.raw = raw
        self.fields = fields

        self.entities = []
        self.included_keys = []

        for obj in self.objs:
            if self.has_pk:
                # We must convert the PK value here, even though this normally happens in
                # django_instance_to_entities otherwise
                # custom PK fields don't work properly
                value = self.model._meta.pk.get_db_prep_save(self.model._meta.pk.pre_save(obj, True), self.connection)
                self.included_keys.append(get_datastore_key(self.connection, self.model, value) if value else None)

                if value == 0:
                    raise IntegrityError("The datastore doesn't support 0 as a key value")

                if not self.model._meta.pk.blank and self.included_keys[-1] is None:
                    raise IntegrityError("You must specify a primary key value for {} instances".format(self.model))
            else:
                # We zip() self.entities and self.included_keys in execute(), so they should be the same length
                self.included_keys.append(None)

            # We don't use the values returned, but this does make sure we're
            # doing the same validation as Django. See issue #493 for an
            # example of how not doing this can mess things up
            for field in fields:
                field.get_db_prep_save(
                    getattr(obj, field.attname) if raw else field.pre_save(obj, True), connection=connection
                )

            primary, descendents = django_instance_to_entities(self.connection, self.fields, self.raw, obj)

            # Append the entity, and any descendents to the list to insert
            self.entities.append((primary, descendents))

    def execute(self):
        """
        Returns the keys of all entities succesfully put into the datastore.

        Under the hood this handles a few implementation specific details,
        such as checking that any unique constraints defined on the entity
        model are respected.
        """
        check_existence = self.has_pk and not has_concrete_parents(self.model)

        def perform_insert(entities):
            results = []
            rpc = transaction._rpc(self.connection.alias)
            must_handle_unique = has_active_unique_constraints(self.model)

            for primary, descendents in entities:
                if primary.key.is_partial:
                    primary.key = primary.key.completed_key(
                        rpc._generate_id()
                    )

                # thanks to cloud firestore in datastore mode strong consistency
                # we can query for the relevant entities to enforce uniqueness
                if must_handle_unique:
                    perform_unique_checks(
                        self.namespace,
                        self.model,
                        rpc,
                        primary,
                        test_fn=lambda stored: stored and len(stored) > 0
                    )

                rpc.put(primary)
                new_key = primary.key

                if descendents:
                    for i, descendent in enumerate(descendents):
                        key = rpc.key(
                            descendent.kind, descendent.key.id_or_name, parent=new_key
                        )
                        descendents[i] = Entity(key)
                        descendents[i].update(descendent)

                    for descendent in descendents:
                        rpc.put(descendent)

                results.append(new_key)
            return results

        @transaction.atomic()
        def insert_chunk(keys, entities):
            for key in keys:
                # sanity check the key isn't already taken
                if check_existence and key is not None:
                    if utils.key_exists(self.connection.alias, key):
                        raise IntegrityError("Tried to INSERT with existing key")

                    # quick validation of the ID value
                    id_or_name = key.id_or_name
                    if isinstance(id_or_name, str) and id_or_name.startswith("__"):
                        raise NotSupportedError("Datastore ids cannot start with __. Id was {}".format(id_or_name))

                    # notify App Engine of any keys we're specifying intentionally
                    reserve_id(self.connection, key.kind, key.id_or_name, self.namespace)

            results = perform_insert(entities)

            if has_active_unique_constraints(self.model):
                # if we're doing a bulk insert, due to the isolation of the
                # datastore inside transactions we won't find duplicate entities
                # to avoid this we can do an in memory comparison instead
                if len(entities) > 1:
                    check_unique_markers_in_memory(self.model, entities)

            caching.add_entities_to_cache(
                self.model,
                [x[0] for x in entities],
                caching.CachingSituation.DATASTORE_GET_PUT,
                self.namespace
            )

            return results

        return insert_chunk(self.included_keys, self.entities)

    def lower(self):
        """
            This exists solely for django-debug-toolbar compatibility.
        """
        return str(self).lower()

    def __str__(self):
        return generate_sql_representation(self)


class DeleteCommand(object):
    """
    Delete an entity / multiple entities.

    Limits imposed by the Firestore in Datastore mode (such as 500 write operations
    per batch) and the backend internal implementation details are handled under the hood.
    """

    def __init__(self, connection, query):
        self.connection = connection
        self.model = query.model
        self.namespace = connection.namespace

        self.select = SelectCommand(connection, query, keys_only=True)
        self.query = self.select.query  # we only need this for the generate_sql_formatter caller...

        # It seems query.tables is populated in most cases, but I have seen cases (albeit in testing)
        # where this isn't the case (particularly when not filtering on anything). In that case
        # fallback to the model table (perhaps we should do
        try:
            table = query.tables[0]
        except (AttributeError, IndexError):
            table = utils.get_top_concrete_parent(query.model)._meta.db_table
        self.table_to_delete = table  # used in wipe_polymodel_from_entity

    def execute(self):
        """
            Ideally we'd just be able to tell appengine to delete all the entities
            which match the query, that would be nice wouldn't it?

            Except we can't. Firstly delete() only accepts keys so we first have to
            execute a keys_only query to find the entities that match the query, then send
            those keys to delete().

            And then there are polymodels (model inheritence) which means we might not even be
            deleting the entity after all, only deleting some of the fields from it.

            What we do then is do a keys_only query, then iterate the entities in batches of
            500, each entity in the batch has its polymodel fields wiped out
            (if necessary) and then we do either a put() or delete() all inside a transaction.

            Oh, and we wipe out memcache in an independent transaction.

            Things to improve:

            - Check the entity matches the query still (there's a fixme there)
        """
        from .indexing import indexers_for_model

        @transaction.atomic()
        def delete_batch(key_slice):
            """
                Batch fetch entities, wiping out any polymodel fields if
                necessary, before deleting the entities by key.

                Any memcache references are also removed.
            """
            entities_to_delete = []
            entities_to_update = []
            updated_keys = []

            # get() expects Key objects, not just dicts with id keys
            keys_in_slice = [get_datastore_key(self.connection, self.model, key_id) for key_id in key_slice]
            entities = transaction._rpc(self.connection.alias).get(keys_in_slice)
            for entity in entities:

                # make sure the entity still exists
                if entity is None:
                    continue

                # handle polymodels
                _wipe_polymodel_from_entity(entity, self.table_to_delete)

                if not entity.get("class"):
                    entities_to_delete.append(entity)
                else:
                    entities_to_update.append(entity)
                updated_keys.append(entity)

            # we don't need an explicit batch here, as we are inside a transaction
            # which already applies this behaviour of non blocking RPCs until
            # the transaction is commited

            client = transaction._rpc(self.connection.alias)

            for entity in entities_to_delete:
                client.delete(entity.key)

            for entity in entities_to_update:
                client.put(entity)

            # Clean up any special indexes that need to be removed
            for indexer in indexers_for_model(self.model):
                for entity in entities_to_delete:
                    indexer.cleanup(client, entity.key)

            # Remove any cache keys
            remove_entities_from_cache_by_key(updated_keys, self.namespace)

            return len(updated_keys)

        # grab the result of the keys only query (see __init__)
        self.select.execute()
        key_ids = [x[self.model._meta.pk.name] for x in self.select.results]

        # for now we can only process 500 entities
        # otherwise we need to handle rollback of independent transactions
        # and race conditions between items being deleted and restored...
        max_batch_size = transaction.TRANSACTION_ENTITY_LIMIT

        if len(key_ids) > max_batch_size:
            raise BulkDeleteError(
                "Bulk deletes for {} can only delete {} instances per batch".format(self.model, max_batch_size)
            )

        return delete_batch(key_ids)

    def lower(self):
        """
            This exists solely for django-debug-toolbar compatibility.
        """
        return str(self).lower()

    def __str__(self):
        return generate_sql_representation(self)


def _wipe_polymodel_from_entity(entity, db_table):
    """
        Wipes out the fields associated with the specified polymodel table.
    """
    polymodel_value = entity.get("class", [])
    if polymodel_value and db_table in polymodel_value:
        # Remove any local fields from this model from the entity
        model = utils.get_model_from_db_table(db_table)
        for field in model._meta.local_fields:
            col = field.column
            if col in entity:
                del entity[col]

        # Then remove this model from the polymodel heirarchy
        polymodel_value.remove(db_table)
        if polymodel_value:
            entity["class"] = polymodel_value


class UpdateCommand(object):
    def __init__(self, connection, query):
        self.model = query.model
        self.select = SelectCommand(connection, query, keys_only=True)
        self.query = self.select.query
        self.values = query.values
        self.connection = connection
        self.namespace = connection.namespace
        self.results = []

    def __str__(self):
        return generate_sql_representation(self)

    def lower(self):
        """
            This exists solely for django-debug-toolbar compatibility.
        """
        return str(self).lower()

    def _update_entity(self, key):
        def update_txt():
            result = transaction._rpc(self.connection.alias).get(key)
            if result is None:
                # Return false to indicate update failure
                return False

            original = copy.deepcopy(result)

            instance_kwargs = {field.attname: value for field, param, value in self.values}

            # Note: If you replace MockInstance with self.model, you'll find that some delete
            # tests fail in the test app. This is because any unspecified fields would then call
            # get_default (even though we aren't going to use them) which may run a query which
            # fails inside this transaction. Given as we are just using MockInstance so that we can
            # call django_instance_to_entities it on it with the subset of fields we pass in,
            # what we have is fine.
            meta = self.model._meta
            instance = MockInstance(_original=MockInstance(_meta=meta, **result), _meta=meta, **instance_kwargs)

            # Convert the instance to an entity
            primary, descendents = django_instance_to_entities(
                self.connection,
                [x[0] for x in self.values],  # Pass in the fields that were updated
                True,
                instance,
                model=self.model,
            )

            # Update the entity we read above with the new values
            result.update(primary)

            # Remove fields which have been marked to be unindexed
            for col in getattr(primary, "_properties_to_remove", []):
                if col in result:
                    del result[col]

            # Make sure that any polymodel classes which were in the original entity are kept,
            # as django_instance_to_entities may have wiped them as well as added them.
            polymodel_classes = list(
                set(original.get(POLYMODEL_CLASS_ATTRIBUTE, []) + result.get(POLYMODEL_CLASS_ATTRIBUTE, []))
            )
            if polymodel_classes:
                result[POLYMODEL_CLASS_ATTRIBUTE] = polymodel_classes

            def perform_insert():
                """
                    Inserts result, and any descendents with their ancestor
                    value set
                """
                client = transaction._rpc(self.connection.alias)

                def test_fn(stored):
                    return stored and len(stored) == 1 and stored[0].key != result.key

                if has_active_unique_constraints(self.model):
                    perform_unique_checks(self.namespace, self.model, client, primary, test_fn)

                inserted_key = client.put(result)
                self.results.append((result, None))

                if descendents:
                    for i, descendent in enumerate(descendents):
                        descendents[i] = Entity(
                            client.key(
                                descendent.kind,
                                descendent.key.name if descendent.key.id is None else descendent.key.id,
                                parent=inserted_key,
                            )
                        )
                        descendents[i].update(descendent)
                    client.put_multi(descendents)

            # this will be async as we're inside a transaction block
            perform_insert()

            # Update the cache with the entity
            caching.add_entities_to_cache(
                self.model,
                [result],
                caching.CachingSituation.DATASTORE_PUT,
                self.namespace
            )

            # Return true to indicate update success
            return True

        return update_txt()

    def execute(self):
        must_handle_unique = has_active_unique_constraints(self.model)

        # TODO: potential optimisation - only use transaction if updating things
        # that require unique checks. This could potentially open a can of worms
        # but would have the benefit of removing any per-commit limits the datastore
        # currently imposes on transactions
        @transaction.atomic()
        def perform_update(results):
            i = 0
            for result in results:
                if self._update_entity(result.key):
                    i += 1

            # due to the isolation of the datastore inside transactions we won't
            # find unique clashes as part of the bulk operation. To avoid this
            # we can do an in memory comparison on the entities themselves
            if must_handle_unique:
                check_unique_markers_in_memory(self.model, self.results)

            return i

        self.select.execute()
        results = list(self.select.results)

        # optimisation: if update is being performed on more than 1 entity,
        # check that the fields in the update are not, on their own, one
        # of the unique constraint combination - otherwise, it means a unique
        # constraint would be violated and there would be no point in continuing
        if len(results) > 1 and must_handle_unique:
            combinations = _unique_combinations(self.model, ignore_pk=True)
            updating_attributes = set([field.attname for field, _, _ in self.values])
            for combination in combinations:
                if set(combination) == updating_attributes:
                    raise IntegrityError('UPDATE on {} field(s) violates unique constraint'.format(combination))

        return perform_update(results)



from django.db import NotSupportedError
from django.db.models.sql.datastructures import EmptyResultSet
from django.db.models.query import QuerySet

try:
    from django.db.models.query import FlatValuesListIterable
except ImportError:
    # Django 1.8
    class FlatValuesListIterable(object):
        pass


try:
    from django.db.models.query import ValuesListQuerySet
except ImportError:
    # Django >= 1.9
    class ValuesListQuerySet(object):
        pass


from .base import BaseParser
from ..query import WhereNode


from djangae.db.utils import get_top_concrete_parent


class Parser(BaseParser):
    def _where_node_leaf_callback(self, node, negated, new_parent, connection, model, compiler):
        new_node = WhereNode(new_parent.using)

        if not hasattr(node, "lhs"):
            raise NotSupportedError("Attempted probable subquery, these aren't supported on the datastore")

        # Leaf
        if hasattr(node.lhs, "target"):
            # from Django 1.9, some node.lhs might not have a target attribute
            # as they might be wrapping date fields
            field = node.lhs.target
            operator = node.lookup_name
        else:
            field = node.lhs.lhs.target
            operator = node.lhs.lookup_name

            # This deals with things like datefield__month__gt=X which means from this point
            # on, operator will have two parts in that particular case and will probably need to
            # be dealt with by a special indexer
            if node.lookup_name != operator:
                operator = "{}__{}".format(operator, node.lookup_name)

        # FIXME: This is a hack. In the base parser (> 1.9) we make use of the get_rhs_op
        # method on lookups to work out what these operators should be
        if field.db_type(connection) in ("set", "list"):
            if operator == "isempty":
                operator = "isnull"
            elif operator == "overlap":
                operator = "in"
            elif operator == "contains":
                operator = "exact"

        if get_top_concrete_parent(field.model) != get_top_concrete_parent(model):
            raise NotSupportedError("Cross-join where filters are not supported on the datastore")

        # Make sure we don't let people try to filter on a text field, otherwise they just won't
        # get any results!

        lookup_supports_text = getattr(node, "lookup_supports_text", False)

        if field.db_type(connection) in ("bytes", "text") and not lookup_supports_text:
            raise NotSupportedError("You can't filter on text or blob fields on the datastore")

        if operator == "isnull" and field.model._meta.parents.values():
            raise NotSupportedError("isnull lookups on inherited relations aren't supported on the datastore")

        lhs = field.column

        try:
            if hasattr(node.rhs, "get_compiler"):
                # This is a subquery
                raise NotSupportedError("Attempted to run a subquery on the datastore")
            elif isinstance(node.rhs, ValuesListQuerySet):
                # We explicitly handle ValuesListQuerySet because of the
                # common case of pk__in=Something.objects.values_list("pk", flat=True)
                # this WILL execute another query, but that is to be expected on a
                # non-relational datastore.

                node.rhs = [x for x in node.rhs]  # Evaluate the queryset
                rhs = node.process_rhs(None, connection)  # Process the RHS as if it was a list

            elif isinstance(node.rhs, QuerySet):
                # In Django 1.9, ValuesListQuerySet doesn't exist anymore, and instead
                # values_list returns a QuerySet
                if node.rhs._iterable_class == FlatValuesListIterable:
                    # if the queryset has FlatValuesListIterable as iterable class
                    # then it's a flat list, and we just need to evaluate the
                    # queryset converting it into a list
                    node.rhs = [x for x in node.rhs]
                else:
                    # otherwise, we try to get the PK from the queryset
                    node.rhs = list(node.rhs.values_list("pk", flat=True))

                rhs = node.process_rhs(None, connection)  # Process the RHS as if it was a list

            else:
                rhs = node.process_rhs(None, connection)
        except EmptyResultSet:
            if operator == "in":
                # Deal with this later
                rhs = [[]]
            else:
                raise

        if operator in ("in", "range"):
            rhs = rhs[-1]
        elif operator == "isnull":
            rhs = node.rhs
        else:
            rhs = rhs[-1][0]

        new_node.set_leaf(
            lhs,
            "__".join([operator, node.path]) if hasattr(node, "path") else operator,
            rhs,
            is_pk_field=field == model._meta.pk,
            negated=negated,
            lookup_name=node.lookup_name,
            namespace=connection.namespace,
            target_field=field,
        )

        # For some reason, this test:
        # test_update_with_related_manager (get_or_create.tests.UpdateOrCreateTests)
        # ends up with duplicate nodes in the where tree. I don't know why. But this
        # weirdly causes the datastore query to return nothing.
        # so here we don't add duplicate nodes, I can't think of a case where that would
        # change the query if it's under the same parent.
        if new_node in new_parent.children:
            return

        new_parent.children.append(new_node)



from collections import OrderedDict

from django.db import (
    connection,
    models,
)
from gcloudc.db.models.fields.json import JSONField
from google.cloud.datastore.entity import Entity

from . import TestCase


class JSONFieldModel(models.Model):
    json_field = JSONField(use_ordered_dict=True, blank=True)


class JSONFieldWithDefaultModel(models.Model):
    json_field = JSONField(use_ordered_dict=True)


class JSONFieldModelTests(TestCase):
    def test_invalid_data_in_datastore_doesnt_throw_an_error(self):
        """
            If invalid data is found while reading the entity data, then
            we should silently ignore the error and just return the data as-is
            rather than converting to list/dict.
            The reason is that if we blow up on load, then there's no way to load the
            entity (in Django) to repair the data. This is also consistent with the behaviour
            of Django when (for example) you load a NULL from the database into a field that is
            non-nullable. The field value will still be None when read.
        """
        bool(JSONFieldModel.objects.exists())  # Force a connection (Django might not have made one yet)

        client = connection.connection.gclient
        entity = Entity(
            client.key(JSONFieldModel._meta.db_table, 1, namespace=connection.settings_dict["NAMESPACE"])
        )
        entity["json_field"] = "bananas"
        client.put(entity)

        instance = JSONFieldModel.objects.get(pk=1)
        self.assertEqual(instance.json_field, "bananas")

    def test_object_pairs_hook_with_ordereddict(self):
        items = [("first", 1), ("second", 2), ("third", 3), ("fourth", 4)]
        od = OrderedDict(items)

        thing = JSONFieldModel(json_field=od)
        thing.save()

        thing = JSONFieldModel.objects.get()
        self.assertEqual(od, thing.json_field)

    def test_object_pairs_hook_with_normal_dict(self):
        """
        Check that dict is not stored as OrderedDict if
        object_pairs_hook is not set
        """

        # monkey patch field
        field = JSONFieldModel._meta.get_field("json_field")
        field.use_ordered_dict = False

        normal_dict = {"a": 1, "b": 2, "c": 3}

        thing = JSONFieldModel(json_field=normal_dict)
        self.assertFalse(isinstance(thing.json_field, OrderedDict))
        thing.save()

        thing = JSONFieldModel.objects.get()
        self.assertFalse(isinstance(thing.json_field, OrderedDict))

        field.use_ordered_dict = True

    def test_float_values(self):
        """ Tests that float values in JSONFields are correctly serialized over repeated saves.
            Regression test for 46e685d4, which fixes floats being returned as strings after a second save.
        """
        test_instance = JSONFieldModel(json_field={"test": 0.1})
        test_instance.save()

        test_instance = JSONFieldModel.objects.get()
        test_instance.save()

        test_instance = JSONFieldModel.objects.get()
        self.assertEqual(test_instance.json_field["test"], 0.1)

    def test_defaults_are_handled_as_pythonic_data_structures(self):
        """ Tests that default values are handled as python data structures and
            not as strings. This seems to be a regression after changes were
            made to remove Subfield from the JSONField and simply use TextField
            instead.
        """
        thing = JSONFieldModel()
        self.assertEqual(thing.json_field, {})

    def test_default_value_correctly_handled_as_data_structure(self):
        """ Test that default value - if provided is not transformed into
            string anymore. Previously we needed string, since we used
            SubfieldBase in JSONField. Since it is now deprecated we need
            to change handling of default value.
        """
        thing = JSONFieldWithDefaultModel()
        self.assertEqual(thing.json_field, {})



from django.db import NotSupportedError
from django.db.models.aggregates import Aggregate
from django.db.models.expressions import (
    Expression,
    Col,
    F,
)

from gcloudc.db.backends.datastore.utils import get_prepared_db_value

CONNECTORS = {
    F.ADD: lambda l, r: l + r,
    F.SUB: lambda l, r: l - r,
    F.MUL: lambda l, r: l * r,
    F.DIV: lambda l, r: l / r,
}


def evaluate_expression(expression, instance, connection):
    """ A limited evaluator for Django's F expressions. This are evaluated within
        the get/put transaction in _update_entity so these will happen atomically
    """

    if isinstance(expression, (str, int, float)):
        return expression

    if isinstance(expression, Aggregate):
        raise NotSupportedError("Aggregate expressions are not supported on the datastore")

    if hasattr(expression, "name"):
        field = instance._meta.get_field(expression.name)
        return get_prepared_db_value(connection, instance._original, field)

    if isinstance(expression, Col):
        field = expression.field
        return get_prepared_db_value(connection, instance._original, field)

    if hasattr(expression, "value"):
        return expression.value

    if hasattr(expression, "connector") and expression.connector in CONNECTORS:
        if hasattr(expression, "children"):
            lhs, rhs = expression.children
        else:
            lhs, rhs = expression.lhs, expression.rhs

        return CONNECTORS[expression.connector](
            evaluate_expression(lhs, instance, connection), evaluate_expression(rhs, instance, connection)
        )

    raise NotImplementedError("Support for expression %r isn't implemented", expression)


class Scatter(Expression):
    def as_sql(self, compiler, connection, **extra_context):
        return "?", ''



import base64
import copy
import json

from django import forms
from django.apps import apps
from django.contrib import admin

try:
    from django.templatetags.static import static  # Django 3+
except ImportError:
    # Django 2.x
    from django.contrib.admin.templatetags.admin_static import static

from django.db import models
from django.forms.fields import MultipleChoiceField
from django.forms.widgets import SelectMultiple
from django.urls import reverse
from django.utils.encoding import force_text
from django.utils.safestring import mark_safe

from gcloudc.utils import memoized


class TrueOrNullFormField(forms.BooleanField):
    def clean(self, value):
        if value:
            return True
        return None


class ListWidget(forms.TextInput):
    """ A widget for being able to display a ListField and SetField. """

    def render(self, name, value, attrs=None):
        if isinstance(value, (list, tuple, set)):
            value = ", ".join([str(v) for v in value])
        return super(ListWidget, self).render(name, value, attrs)

    def value_from_datadict(self, data, files, name):
        """ Given a dictionary of data and this widget's name, returns the value
            of this widget. Returns None if it's not provided.
        """
        value = data.get(name, "")

        if value is None:
            return None

        if isinstance(value, str):
            value = value.split(",")
        return [v.strip() for v in value if v.strip()]


class ListFormField(forms.Field):
    """ A form field for being able to display a ListField and SetField. """

    widget = ListWidget
    delimiter = ","

    def clean(self, value):
        if value:
            if isinstance(value, (list, tuple, set)):
                self._check_values_against_delimiter(value)
                return value
            return [v.strip() for v in value.split(",") if v.strip()]
        return []

    def _check_values_against_delimiter(self, values):
        delimiter = self.delimiter  # faster
        for value in values:
            assert delimiter not in value


class SetSelectMultipleWidget(SelectMultiple):
    def format_value(self, value):
        """Return selected values as a list."""
        # the default implementation in `SelectMultiple` does not pass `set` to
        # the `isinstance` check, which generates a list containing one string
        # for the initial value, rather than a list of multiple strings
        if not isinstance(value, (tuple, list, set)):
            value = [value]
        return [force_text(v) if v is not None else "" for v in value]


class SetMultipleChoiceField(MultipleChoiceField):
    """A form field to handle the initial values for a set field with multiple choices."""

    widget = SetSelectMultipleWidget


class JSONWidget(forms.Textarea):
    """ A widget for being able to display a JSONField in a form. """

    def render(self, name, value, attrs=None):
        """ Dump the python object to JSON if it hasn't been done yet. """
        from django.core.serializers.json import DjangoJSONEncoder

        if not isinstance(value, str):
            value = DjangoJSONEncoder().encode(value)
        return super(JSONWidget, self).render(name, value, attrs)


class JSONFormField(forms.CharField):
    """ A form field for being able to display a JSONField in a form.
        The JSON is rendered as string in a textarea, but is parsed to python (be)for validation.
     """

    widget = JSONWidget

    def clean(self, value):
        """ (Try to) parse JSON string back to python. """
        assert isinstance(value, str) or value is None, "JSONField value must be a string or None"

        value = super(JSONFormField, self).clean(value)

        if not value:
            value = None

        if value:
            try:
                value = json.loads(value)
                if not value and self.required:
                    raise forms.ValidationError("Non-empty JSON object is required")

            except ValueError:
                raise forms.ValidationError("Could not parse value as JSON")
        return value


class OrderedModelMultipleChoiceField(forms.ModelMultipleChoiceField):
    def clean(self, value):
        """
        Maintain the order of the values passed in. Without this special casing,
        _check_values() runs a pk__in query which under the hood gets
        ordered by Djangae via its default model ordering (which has a PK
        fallback if no ordering is explicit), and thus the original order of the
        values passed in is lost.
        """
        # Make a copy of the value - we still run it via the normal clean
        # so that validators are run - but we don't return the queryset like
        # the vanilla ModelMultipleChoiceField as the ordering will be lost by
        # doing that, so instead we return a list of the PK values, which is
        # not strictly what we should do, but RelatedListField accepts it and
        # so it makes this work
        value_copy = copy.deepcopy(value)
        super(OrderedModelMultipleChoiceField, self).clean(value_copy)
        return [self.queryset.model._meta.pk.to_python(v) for v in value]


# Basic obfuscation, just so that the db_table doesn't
# appear in forms. (Not the end of the world if it does, but it's nice to
# hide these things). We don't encrypt for performance reasons.
# http://en.wikipedia.org/wiki/Vigen%C3%A8re_cipher

_VC_KEY = "1K94KG8L"  # Fixed key, don't change this!!


def model_path(obj):
    return obj._meta.db_table


def vc_encode(string):
    enc = []
    for i in range(len(string)):
        key_c = _VC_KEY[i % len(_VC_KEY)]
        enc_c = chr((ord(string[i]) + ord(key_c)) % 256)
        enc.append(enc_c)
    ret = base64.urlsafe_b64encode("".join(enc).encode("utf-8")).decode("ascii")
    return ret


def vc_decode(enc):
    dec = []
    enc = base64.urlsafe_b64decode(enc).decode("utf-8")
    for i in range(len(enc)):
        key_c = _VC_KEY[i % len(_VC_KEY)]
        dec_c = chr((256 + ord(enc[i]) - ord(key_c)) % 256)
        dec.append(dec_c)
    return "".join(dec)


def encode_pk(pk, obj_cls):
    assert pk is not None

    return vc_encode("{0}|{1}".format(obj_cls._meta.db_table, pk))


def decode_pk(encoded):
    result = vc_decode(encoded)

    model_ref, pk = result.split("|")
    return model_ref, pk


class GenericRelationWidget(forms.MultiWidget):
    def __init__(self, widgets=None, *args, **kwargs):
        widgets = (forms.Select(), GenericFKInput())
        super(GenericRelationWidget, self).__init__(widgets=widgets, *args, **kwargs)

    def decompress(self, value):
        if isinstance(value, str):
            return decode_pk(value)

        if value:
            return [value._meta.db_table, value.pk]

        return [None, None]


class GenericFKInput(forms.TextInput):
    def __init__(self, *args, **kwargs):
        super(GenericFKInput, self).__init__(*args, **kwargs)

    def render(self, name, value, attrs):
        urls = {}
        for m in admin.site._registry:
            urls[model_path(m)] = reverse("admin:%s_%s_changelist" % (m._meta.app_label, m._meta.model_name))
        urls = json.dumps(urls)
        safe_name = name.replace("-", "_")
        extra = []
        extra.append(
            u"""<script type="text/javascript">var urls = %s; function popup_%s(trigger) {
            var name = trigger.id.replace(/^lookup_/, '');
            name = id_to_windowname(name);
            var chosen_model = django.jQuery(trigger).siblings('select').val();
            if (chosen_model in urls) {
                var href = urls[chosen_model] + '?pop=1'
                var win = window.open(href, name, 'height=500,width=800,resizable=yes,scrollbars=yes');
                win.focus();
            }
            return false;
            }
        </script>\n"""
            % (urls, safe_name)
        )
        extra.append(
            u'<a href="" class="related-lookup" id="lookup_id_%s" onclick="return popup_%s(this);"> '
            % (name, safe_name)
        )
        extra.append(
            u'<img src="%s" width="16" height="16" alt="%s" /></a>'
            % (static("admin/img/selector-search.gif"), "Lookup")
        )
        output = [super(GenericFKInput, self).render(name, value, attrs)] + extra
        return mark_safe(u"".join(output))


@memoized
def model_from_db_table(db_table):
    for model in apps.get_models():
        if model._meta.db_table == db_table:
            return model
    raise ValueError("Couldn't find model class for %s" % db_table)


_CHOICES = None


@memoized
def get_all_model_choices():
    global _CHOICES
    if _CHOICES is None:
        _CHOICES = [("", "None")] + [(model_path(m), m.__name__) for m in apps.get_models()]
    return _CHOICES


class GenericRelationFormfield(forms.MultiValueField):
    def __init__(self, fields=None, widget=None, choices=None, *args, **kwargs):
        fields = (forms.ChoiceField(), forms.CharField(max_length=30))

        super(GenericRelationFormfield, self).__init__(fields=fields, widget=GenericRelationWidget(), *args, **kwargs)
        self.widget.widgets[0].choices = self.fields[0].choices = choices or get_all_model_choices()

    @classmethod
    def to_python(cls, value, model_ref=None, pk=None):
        if model_ref is None:
            if value is None:
                return None

            if isinstance(value, models.Model):
                return value

            model_ref, pk = decode_pk(value)

        try:
            pk = int(pk)
        except (ValueError, TypeError):
            raise forms.ValidationError("Invalid instance key.")

        model = cls.load_model(model_ref)
        try:
            return model.objects.get(pk=pk)
        except model.DoesNotExist:
            raise forms.ValidationError("Invalid instance key.")

    @classmethod
    def load_model(cls, model_ref):
        return model_from_db_table(model_ref)

    def validate(self, value):
        self.to_python(value)

    @classmethod
    def to_string(cls, value):
        if value is None:
            return None

        if isinstance(value, str):
            return value

        return encode_pk(value.pk, value)

    def compress(self, data_list):
        if not data_list or not data_list[0]:
            return None
        return self.to_python(None, *data_list)



from datetime import datetime
from decimal import Decimal
from itertools import chain

from django.apps import apps
from django.conf import settings
from django.db import IntegrityError
from django.db.backends.utils import format_number
from django.utils import timezone
from google.cloud.datastore.entity import Entity
from google.cloud.datastore.key import Key

from gcloudc.utils import memoized

from .query_utils import get_filter

try:
    from django.db.models.expressions import BaseExpression
except ImportError:
    from django.db.models.expressions import ExpressionNode as BaseExpression


def make_utc_compatible(value):
    if value is None:
        return None

    if timezone.is_aware(value):
        if settings.USE_TZ:
            value = value.astimezone(timezone.utc)
        else:
            raise ValueError("Djangae backend does not support timezone-aware datetimes when USE_TZ is False.")
    return value


@memoized
def get_model_from_db_table(db_table):
    # We use include_swapped=True because tests might need access to gauth User models which are
    # swapped if the user has a different custom user model

    kwargs = {"include_auto_created": True, "include_swapped": True}

    for model in apps.get_models(**kwargs):
        if model._meta.db_table == db_table:
            return model


def decimal_to_string(value, max_digits=16, decimal_places=0):
    """
    Converts decimal to a unicode string for storage / lookup by nonrel
    databases that don't support decimals natively.

    This is an extension to `django.db.backends.utils.format_number`
    that preserves order -- if one decimal is less than another, their
    string representations should compare the same (as strings).

    TODO: Can't this be done using string.format()?
          Not in Python 2.5, str.format is backported to 2.6 only.
    """

    # Handle sign separately.
    if value.is_signed():
        sign = u"-"
        value = abs(value)
    else:
        sign = u""

    # Let Django quantize and cast to a string.
    value = format_number(value, max_digits, decimal_places)

    # Pad with zeroes to a constant width.
    n = value.find(".")
    if n < 0:
        n = len(value)
    if n < max_digits - decimal_places:
        value = u"0" * (max_digits - decimal_places - n) + value
    return sign + value


def normalise_field_value(value):
    """ Converts a field value to a common type/format to make comparable to another. """
    if isinstance(value, datetime):
        return make_utc_compatible(value)
    elif isinstance(value, Decimal):
        return decimal_to_string(value)
    return value


def get_datastore_kind(model):
    return get_top_concrete_parent(model)._meta.db_table


def get_prepared_db_value(connection, instance, field, raw=False):
    value = getattr(instance, field.attname) if raw else field.pre_save(instance, instance._state.adding)

    if isinstance(value, BaseExpression):
        from gcloudc.db.backends.datastore.expressions import evaluate_expression

        # We can't actually support F expressions on the datastore, but we can simulate
        # them, evaluating the expression in place.

        # TODO: For saves and updates we should raise a Warning. When evaluated in a filter
        # we should raise an Error
        value = evaluate_expression(value, instance, connection)

    if hasattr(value, "prepare_database_save"):
        value = value.prepare_database_save(field)
    else:
        value = field.get_db_prep_save(value, connection=connection)

    value = connection.ops.value_for_db(value, field)

    return value


def get_concrete_parents(model, ignore_leaf=False):
    ret = [x for x in model.mro() if hasattr(x, "_meta") and not x._meta.abstract and not x._meta.proxy]
    if ignore_leaf:
        ret = [x for x in ret if x != model]
    return ret


@memoized
def get_top_concrete_parent(model):
    return get_concrete_parents(model)[-1]


def get_concrete_fields(model, ignore_leaf=False):
    """
        Returns all the concrete fields for the model, including those
        from parent models
    """
    concrete_classes = get_concrete_parents(model, ignore_leaf)
    fields = []
    for klass in concrete_classes:
        fields.extend(klass._meta.fields)

    return fields


@memoized
def get_concrete_db_tables(model):
    return [x._meta.db_table for x in get_concrete_parents(model)]


@memoized
def has_concrete_parents(model):
    return get_concrete_parents(model) != [model]


@memoized
def get_field_from_column(model, column):
    for field in model._meta.fields:
        if field.column == column:
            return field
    return None


def django_instance_to_entities(connection, fields, raw, instance, check_null=True, model=None):
    """
        Converts a Django Model instance to an App Engine `Entity`

        Arguments:
            connection: Djangae appengine connection object
            fields: A list of fields to populate in the Entity
            raw: raw flag to pass to get_prepared_db_value
            instance: The Django model instance to convert
            check_null: Whether or not we should enforce NULL during conversion
            (throws an error if None is set on a non-nullable field)
            model: Model class to use instead of the instance one

        Returns:
            entity, [entity, entity, ...]

       Where the first result in the tuple is the primary entity, and the
       remaining entities are optionally descendents of the primary entity. This
       is useful for special indexes (e.g. contains)
    """

    from gcloudc.db.backends.datastore.indexing import special_indexes_for_column, get_indexer, IgnoreForIndexing
    from gcloudc.db.backends.datastore import POLYMODEL_CLASS_ATTRIBUTE

    model = model or type(instance)
    inheritance_root = get_top_concrete_parent(model)

    db_table = get_datastore_kind(inheritance_root)

    def value_from_instance(_instance, _field):
        value = get_prepared_db_value(connection, _instance, _field, raw)

        # If value is None, but there is a default, and the field is not nullable then we should populate it
        # Otherwise thing get hairy when you add new fields to models
        if value is None and _field.has_default() and not _field.null:
            # We need to pass the default through get_db_prep_save to properly do the conversion
            # this is how
            value = _field.get_db_prep_save(_field.get_default(), connection)

        if check_null and (not _field.null and not _field.primary_key) and value is None:
            raise IntegrityError("You can't set %s (a non-nullable field) to None!" % _field.name)

        is_primary_key = False
        if _field.primary_key and _field.model == inheritance_root:
            is_primary_key = True

        return value, is_primary_key

    field_values = {}
    primary_key = None

    descendents = []
    fields_to_unindex = set()

    for field in fields:
        value, is_primary_key = value_from_instance(instance, field)
        if is_primary_key:
            primary_key = value
        else:
            field_values[field.column] = value

        # Add special indexed fields
        for index in special_indexes_for_column(model, field.column):
            indexer = get_indexer(field, index)

            unindex = False
            try:
                values = indexer.prep_value_for_database(
                    value, index, model=model, column=field.column, connection=connection
                )
            except IgnoreForIndexing as e:
                # We mark this value as being wiped out for indexing
                unindex = True
                values = e.processed_value

            if not hasattr(values, "__iter__") or isinstance(values, (bytes, str)):
                values = [values]

            # If the indexer returns additional entities (instead of indexing a special column)
            # then just store those entities
            if indexer.PREP_VALUE_RETURNS_ENTITIES:
                descendents.extend(values)
            else:
                for i, v in enumerate(values):
                    column = indexer.indexed_column_name(field.column, v, index)

                    if unindex:
                        fields_to_unindex.add(column)
                        continue

                    # If the column already exists in the values, then we convert it to a
                    # list and append the new value
                    if column in field_values:
                        if not isinstance(field_values[column], list):
                            field_values[column] = [field_values[column], v]
                        else:
                            field_values[column].append(v)
                    else:
                        # Otherwise we just set the column to the value
                        field_values[column] = v

    args = [db_table]
    if primary_key is not None:
        args.append(primary_key)

    key = Key(*args, namespace=connection.namespace, project=connection.gcloud_project)

    exclude_from_indexes = tuple(
        field.column for field in model._meta.fields
        if field.db_type(connection) in ('text', 'bytes')
    )

    entity = Entity(key, exclude_from_indexes)
    entity.update(field_values)

    if fields_to_unindex:
        entity._properties_to_remove = fields_to_unindex

    classes = get_concrete_db_tables(model)
    if len(classes) > 1:
        entity[POLYMODEL_CLASS_ATTRIBUTE] = list(set(classes))

    return entity, descendents


def get_datastore_key(connection, model, pk):
    """ Return a datastore.Key for the given model and primary key.
    """
    factory = connection.connection.gclient.key
    kind = get_top_concrete_parent(model)._meta.db_table
    return factory(kind, pk, namespace=connection.namespace)


class MockInstance(object):
    """
        This creates a mock instance for use when passing a datastore entity
        into get_prepared_db_value. This is used when performing updates to prevent a complete
        conversion to a Django instance before writing back the entity
    """

    def __init__(self, **kwargs):
        is_adding = kwargs.pop("_is_adding", False)
        self._original = kwargs.pop("_original", None)
        self._meta = kwargs.pop("_meta", None)

        class State:
            adding = is_adding

        self.fields = {}
        for field_name, value in kwargs.items():
            self.fields[field_name] = value

        self._state = State()

    def __getattr__(self, attr):
        if attr in self.fields:
            return self.fields[attr]
        raise AttributeError(attr)


def key_exists(connection, key):
    from . import transaction
    qry = transaction._rpc(connection).query(namespace=key.namespace, ancestor=key)
    qry.keys_only()
    qry.add_filter("__key__", "=", key)
    return count_query(qry) > 0


# Null-friendly comparison functions


def lt(x, y):
    if x is None and y is None:
        return False
    if x is None and y is not None:
        return True
    elif x is not None and y is None:
        return False
    elif isinstance(x, Key) and isinstance(y, Key):
        lhs = tuple([x.namespace] + list(x.flat_path))
        rhs = tuple([y.namespace] + list(y.flat_path))
        return lhs < rhs
    else:
        return x < y


def gt(x, y):
    if x is None and y is None:
        return False
    if x is None and y is not None:
        return False
    elif x is not None and y is None:
        return True
    elif isinstance(x, Key) and isinstance(y, Key):
        lhs = tuple([x.namespace] + list(x.flat_path))
        rhs = tuple([y.namespace] + list(y.flat_path))
        return lhs > rhs
    else:
        return x > y


def gte(x, y):
    return not lt(x, y)


def lte(x, y):
    return not gt(x, y)


def django_ordering_comparison(ordering, lhs, rhs):
    if not ordering:
        return -1  # Really doesn't matter

    ASCENDING = 1
    DESCENDING = 2

    for order in ordering:
        direction = DESCENDING if order.startswith("-") else ASCENDING
        order = order.lstrip("-")

        if lhs is not None:
            lhs_value = lhs.key if order == "__key__" else lhs.get(order)
        else:
            lhs_value = None

        if rhs is not None:
            rhs_value = rhs.key if order == "__key__" else rhs.get(order)
        else:
            rhs_value = None

        if direction == ASCENDING and lhs_value != rhs_value:
            return -1 if lt(lhs_value, rhs_value) else 1
        elif direction == DESCENDING and lhs_value != rhs_value:
            return 1 if lt(lhs_value, rhs_value) else -1

    return 0


def entity_matches_query(entity, query):
    """
        Return True if the entity would potentially be returned by the datastore
        query
    """
    from . import meta_queries

    OPERATORS = {"=": lambda x, y: x == y, "<": lt, ">": gt, "<=": lte, ">=": gte}

    queries = [query]
    if isinstance(query, meta_queries.AsyncMultiQuery):
        raise NotImplementedError(
            "We just need to separate the multiquery " "into 'queries' then everything should work"
        )

    for query in queries:
        comparisons = chain([("__kind__", "=")], [(x[0], x[1]) for x in query.filters])

        for ent_attr, op in comparisons:
            if ent_attr == "__key__":
                continue

            compare = OPERATORS[op]  # We want this to throw if there's some op we don't know about

            if ent_attr == "__kind__":
                ent_value = entity.kind
                query_value = query.kind
            else:
                query_value = get_filter(query, (ent_attr, op))
                ent_value = entity.get(ent_attr)

            if not isinstance(query_value, (list, tuple)):
                query_values = [query_value]
            else:
                # The query value can be a list of ANDed values
                query_values = query_value

            if not isinstance(ent_value, (list, tuple)):
                ent_value = [ent_value]

            matches = False
            for value in query_values:  # [22, 23]
                # If any of the values don't match then this query doesn't match
                if not any(compare(attr, value) for attr in ent_value):
                    matches = False
                    break
            else:
                # One of the ent_attrs matches the query_attrs
                matches = True

            if not matches:
                # One of the AND values didn't match
                break
        else:
            # If we got through the loop without breaking, then the entity matches
            return True

    return False


def ensure_datetime(value):
    """
        Painfully, sometimes the Datastore returns dates as datetime objects, and sometimes
        it returns them as unix timestamps in microseconds!!
    """
    if isinstance(value, int):
        return datetime.utcfromtimestamp(value / 1e6)
    return value


def count_query(query):
    """
        The Google Cloud Datastore API doesn't expose a way to count a query
        the traditional method of doing a keys-only query is apparently actually
        slower than this method
    """

    # Largest 32 bit number, fairly arbitrary but I've seen Java Cloud Datastore
    # code that uses Integer.MAX_VALUE which is this value
    MAX_INT = 2147483647

    # Setting a limit of zero and an offset of max int will make
    # the server (rather than the client) skip the entities and then
    # return the number of skipped entities, fo realz yo!
    iterator = query.fetch(limit=0, offset=MAX_INT)
    [x for x in iterator]  # Force evaluation of the iterator

    count = iterator._skipped_results
    while iterator._more_results:
        # If we have more results then use cursor offsetting and repeat
        iterator = query.fetch(limit=0, offset=MAX_INT, start_cursor=iterator.next_page_token)
        [x for x in iterator]  # Force evaluation of the iterator

        count += iterator._skipped_results

    return count



# STANDARD LIB
import copy
from decimal import Decimal
from itertools import chain

from django import forms
from django.db import models
from django.db.models.lookups import Lookup, Transform
from django.core.exceptions import ValidationError, ImproperlyConfigured
from django.utils.text import capfirst

from gcloudc.core.validators import MinItemsValidator, MaxItemsValidator
from gcloudc.forms.fields import ListFormField, SetMultipleChoiceField

# types that don't need to be quoted when serializing an iterable field
_SERIALIZABLE_TYPES = (int, float, Decimal)


class _FakeModel(object):
    """
    An object of this class can pass itself off as a model instance
    when used as an arguments to Field.pre_save method (item_fields
    of iterable fields are not actually fields of any model).
    """

    def __init__(self, field, value):
        setattr(self, field.attname, value)


class ContainsLookup(Lookup):
    lookup_name = "contains"

    def get_rhs_op(self, connection, rhs):
        return "= %s" % rhs

    def get_prep_lookup(self):
        if hasattr(self.rhs, "__iter__") and not isinstance(self.rhs, str):
            raise ValueError("__contains cannot take an iterable")

        # Currently, we cannot differentiate between an empty list (which we store as None) and a
        # list which contains None.  Once we move to storing empty lists as empty lists (now that
        # GAE allows it) we can remove this restriction.
        if self.rhs is None:
            raise ValueError("__contains cannot take None, use __isempty instead")

        return self.rhs


class IsEmptyLookup(Lookup):
    lookup_name = "isempty"

    def get_rhs_op(self, connection, rhs):
        return "isnull %s" % rhs

    def get_prep_lookup(self):
        if self.rhs not in (True, False):
            raise ValueError("__isempty takes a boolean as a value")

        return self.rhs


class OverlapLookup(Lookup):
    lookup_name = "overlap"
    get_db_prep_lookup_value_is_iterable = False

    def get_rhs_op(self, connection, rhs):
        return "IN %s" % rhs

    def get_db_prep_lookup(self, value, connection):
        # the In lookup wraps each element in a list, so we unwrap here
        ret = super(OverlapLookup, self).get_db_prep_lookup(value, connection)
        return (ret[0], [x for x in chain(*ret[-1])])

    def get_prep_lookup(self):
        if not isinstance(self.rhs, (list, set)):
            raise ValueError("__overlap takes a list or set as a value")

        return [self.lhs.output_field.get_prep_value(v) for v in self.rhs]


class IterableTransform(Transform):
    lookup_name = "item"

    def __init__(self, item_field_type, *args, **kwargs):
        super(IterableTransform, self).__init__(*args, **kwargs)
        self.item_field_type = item_field_type

    def get_lookup(self, name):
        return self.item_field_type.get_lookup(name)


class IterableTransformFactory(object):
    def __init__(self, base_field):
        self.base_field = base_field

    def __call__(self, *args, **kwargs):
        return IterableTransform(self.base_field, *args, **kwargs)


class IterableField(models.Field):

    choices_form_field_class = forms.MultipleChoiceField

    @property
    def _iterable_type(self):
        raise NotImplementedError()

    def from_db_value(self, value, expression, connection, **kwargs):
        return self.to_python(value)

    def db_type(self, connection):
        return "list"

    def get_lookup(self, name):
        # isnull is explitly not blocked here, because annoyingly Django adds isnull lookups implicitly on
        # excluded nullable with no way of switching it off!

        if name == "exact":
            raise ValueError("You can't perform __{} on an iterable field, did you mean __contains?".format(name))

        if name == "in":
            raise ValueError("You can't perform __{} on an iterable field, did you mean __overlap?".format(name))

        if name in ("regex", "startswith", "endswith", "iexact", "istartswith", "icontains", "iendswith"):
            raise ValueError("You can't perform __{} on an iterable field, did you mean __item__{}?".format(name, name))

        return super(IterableField, self).get_lookup(name)

    def get_transform(self, name):
        if name == "item":
            return IterableTransformFactory(self.item_field_type)

        return super(IterableField, self).get_transform(name)

    def __init__(self, item_field_type, *args, **kwargs):

        # This seems bonkers, we shout at people for specifying null=True, but then do it ourselves. But this is because
        # *we* abuse None values for our own purposes (to represent an empty iterable) if someone else tries to then
        # all hell breaks loose
        if kwargs.get("null", False):
            raise RuntimeError(
                "IterableFields cannot be set as nullable (as the datastore doesn't differentiate None vs []"
            )

        kwargs["null"] = True

        kwargs["default"] = kwargs.get("default", [])

        self._original_item_field_type = copy.deepcopy(item_field_type)  # For deconstruction purposes

        if hasattr(item_field_type, "attname"):
            item_field_type = item_field_type.__class__

        if callable(item_field_type):
            item_field_type = item_field_type()

        if isinstance(item_field_type, models.ForeignKey):
            raise ImproperlyConfigured("Lists of ForeignKeys aren't supported, use RelatedSetField instead")

        self.item_field_type = item_field_type

        # We'll be pretending that item_field is a field of a model
        # with just one "value" field.
        assert not hasattr(self.item_field_type, "attname")
        self.item_field_type.set_attributes_from_name("value")

        # Pop the 'min_length' and 'max_length' from the kwargs, if they're there, as this avoids
        # 'min_length' causing an error when calling super()
        min_length = kwargs.pop("min_length", None)
        max_length = kwargs.pop("max_length", None)

        # Check that if there's a min_length that blank is not True.  This is partly because it
        # doesn't make sense, and partly because if the value (i.e. the list or set) is empty then
        # Django will skip the validators, thereby skipping the min_length check.
        if min_length and kwargs.get("blank"):
            raise ImproperlyConfigured("Setting blank=True and min_length=%d is contradictory." % min_length)

        super(IterableField, self).__init__(*args, **kwargs)

        # Now that self.validators has been set up, we can add the min/max legnth validators
        if min_length is not None:
            self.validators.append(MinItemsValidator(min_length))
        if max_length is not None:
            self.validators.append(MaxItemsValidator(max_length))

    def get_default(self):
        if self.default is None:
            return None
        if callable(self.default):
            return self.default()
        # If the default is an mutable iterable (e.g. list) then we need to make a new instance of
        # it each time we use it
        return self._iterable_type(self.default)

    def deconstruct(self):
        name, path, args, kwargs = super(IterableField, self).deconstruct()
        args = (self._original_item_field_type,)
        del kwargs["null"]
        return name, path, args, kwargs

    def contribute_to_class(self, cls, name):
        self.item_field_type.model = cls
        self.item_field_type.name = name
        super(IterableField, self).contribute_to_class(cls, name)

    def _map(self, function, iterable, *args, **kwargs):
        return self._iterable_type(function(element, *args, **kwargs) for element in iterable)

    def to_python(self, value):
        if value is None:
            return self._iterable_type([])

        # If possible, parse the string into the iterable
        if not hasattr(value, "__iter__") or isinstance(value, str):
            if isinstance(value, str):
                if value.startswith("[") and value.endswith("]"):
                    value = value[1:-1].strip()

                    value = [x.strip("'").strip('"') for x in value.split(",") if len(value) > 2]
                else:
                    raise ValueError("Unable to parse string into iterable field")
            else:
                raise TypeError("Tried to assign non-iterable to an IterableField")

        return self._map(self.item_field_type.to_python, value)

    def pre_save(self, model_instance, add):
        """
            Gets our value from the model_instance and passes its items
            through item_field's pre_save (using a fake model instance).
        """
        value = getattr(model_instance, self.attname)
        if value is None:
            raise ValueError(
                "You can't set a {} to None (did you mean {}?)".format(
                    self.__class__.__name__, str(self._iterable_type())
                )
            )

        if isinstance(value, str):
            # Catch accidentally assigning a string to a ListField
            raise ValueError("Tried to assign a string to a {}".format(self.__class__.__name__))

        return self._map(lambda item: self.item_field_type.pre_save(_FakeModel(self.item_field_type, item), add), value)

    def get_db_prep_value(self, value, connection, prepared=False):
        if not prepared:
            value = self.get_prep_value(value)
            if value is None:
                return None

        # If the value is an empty iterable, store None
        if value == self._iterable_type([]):
            return None

        return self._map(self.item_field_type.get_db_prep_save, value, connection=connection)

    def get_db_prep_lookup(self, lookup_type, value, connection, prepared=False):
        """
        Passes the value through get_db_prep_lookup of item_field.
        """
        return self.item_field_type.get_db_prep_lookup(lookup_type, value, connection=connection, prepared=prepared)

    def validate(self, value_list, model_instance):
        """ We want to override the default validate method from django.db.fields.Field, because it
            is only designed to deal with a single choice from the user.
        """
        if not self.editable:
            # Skip validation for non-editable fields
            return

        # Validate choices
        if self.choices:
            valid_values = []
            for choice in self.choices:
                if isinstance(choice[0], (list, tuple)):
                    # this is an optgroup, so look inside it for the options
                    for optgroup_choice in choice[0]:
                        valid_values.append(optgroup_choice[0])
                else:
                    valid_values.append(choice[0])
            for value in value_list:
                if value not in valid_values:
                    # TODO: if there is more than 1 invalid value then this should show all of the invalid values
                    raise ValidationError(self.error_messages["invalid_choice"] % value)
        # Validate null-ness
        if value_list is None and not self.null:
            raise ValidationError(self.error_messages["null"])

        if not self.blank and not value_list:
            raise ValidationError(self.error_messages["blank"])

        # apply the default items validation rules
        for value in value_list:
            self.item_field_type.clean(value, model_instance)

    def formfield(self, **kwargs):
        """ If this field has choices, then we can use a multiple choice field.
            NB: The choices must be set on *this* field, e.g. this_field = ListField(CharField(), choices=x)
            as opposed to: this_field = ListField(CharField(choices=x))
        """
        # Largely lifted straight from Field.formfield() in django.models.__init__.py
        defaults = {"required": not self.blank, "label": capfirst(self.verbose_name), "help_text": self.help_text}
        if self.has_default():  # No idea what this does
            if callable(self.default):
                defaults["initial"] = self.default
                defaults["show_hidden_initial"] = True
            else:
                defaults["initial"] = self.get_default()

        if self.choices:
            form_field_class = self.choices_form_field_class
            defaults["choices"] = self.get_choices(include_blank=False)  # no empty value on a multi-select
        else:
            form_field_class = ListFormField
        defaults.update(**kwargs)
        return form_field_class(**defaults)

    def value_to_string(self, obj):
        return "[" + ",".join(_serialize_value(o) for o in self.value_from_object(obj)) + "]"


# New API
IterableField.register_lookup(ContainsLookup)
IterableField.register_lookup(OverlapLookup)
IterableField.register_lookup(IsEmptyLookup)


class ListField(IterableField):
    def __init__(self, *args, **kwargs):
        self.ordering = kwargs.pop("ordering", None)
        if self.ordering is not None and not callable(self.ordering):
            raise TypeError("'ordering' has to be a callable or None, " "not of type %r." % type(self.ordering))
        super(ListField, self).__init__(*args, **kwargs)

    def get_internal_type(self):
        return "ListField"

    def pre_save(self, model_instance, add):
        value = super(ListField, self).pre_save(model_instance, add)

        if value and self.ordering:
            value.sort(key=self.ordering)

        return value

    @property
    def _iterable_type(self):
        return list

    def deconstruct(self):
        name, path, args, kwargs = super(ListField, self).deconstruct()
        kwargs["ordering"] = self.ordering
        return name, path, args, kwargs


class SetField(IterableField):

    # specifies a different default widget than the vanilla MultipleChoiceField
    choices_form_field_class = SetMultipleChoiceField

    @property
    def _iterable_type(self):
        return set

    def get_internal_type(self):
        return "SetField"

    def db_type(self, connection):
        return "set"

    def get_db_prep_save(self, *args, **kwargs):
        ret = super(SetField, self).get_db_prep_save(*args, **kwargs)
        if ret:
            ret = list(ret)
        return ret

    def get_db_prep_lookup(self, *args, **kwargs):
        ret = super(SetField, self).get_db_prep_lookup(*args, **kwargs)
        if ret:
            ret = list(ret)
        return ret


def _serialize_value(value):
    if isinstance(value, _SERIALIZABLE_TYPES):
        return str(value)

    if hasattr(value, "isoformat"):
        # handle datetime, date, and time objects
        value = value.isoformat()
    elif not isinstance(value, str):
        value = str(value)

    return "'{0}'".format(value)



"""
    This is a wrapper around the individual backend transactions,
    switching on the connection type.
"""

from django.db import (
    DEFAULT_DB_ALIAS,
    connections,
)
from django.db.transaction import TransactionManagementError
from gcloudc.context_decorator import ContextDecorator
from gcloudc.db.backends.datastore import transaction as datastore_transaction


TransactionManagementError = TransactionManagementError


class Atomic(ContextDecorator):
    # This should be the superset of any connector args (just Datastore for now)
    VALID_ARGUMENTS = datastore_transaction.AtomicDecorator.VALID_ARGUMENTS[:]

    @classmethod
    def _do_enter(cls, state, decorator_args):
        using = decorator_args.get("using", "default") or "default"

        try:
            connections[using]
            state.decorator = datastore_transaction.AtomicDecorator
        except (KeyError, TypeError):
            raise ValueError("Unable to get connection for %s" % using)

        return state.decorator._do_enter(state, decorator_args)

    @classmethod
    def _do_exit(cls, state, decorator_args, exception):
        state.decorator._do_exit(state, decorator_args, exception)


atomic = Atomic


class NonAtomic(ContextDecorator):

    @classmethod
    def _do_enter(cls, state, decorator_args):
        using = decorator_args.get("using", "default") or "default"

        try:
            connections[using]
            state.decorator = datastore_transaction.NonAtomicDecorator
        except (KeyError, TypeError):
            raise ValueError("Unable to get connection for %s" % using)

        return state.decorator._do_enter(state, decorator_args)

    @classmethod
    def _do_exit(cls, state, decorator_args, exception):
        state.decorator._do_exit(state, decorator_args, exception)


non_atomic = NonAtomic


def in_atomic_block(using="default"):
    try:
        connections[using]
        return datastore_transaction.in_atomic_block(using=using)
    except (KeyError, TypeError):
        raise ValueError("Unable to find connection with alias: %s" % using)


def get_connection(using=None):
    """
    Get a database connection by name, or the default database connection
    if no name is provided. This is a private API.
    """
    if using is None:
        using = DEFAULT_DB_ALIAS
    return connections[using]


def on_commit(func, using=None):
    """
    Register `func` to be called when the current transaction is committed.
    If the current transaction is rolled back, `func` will not be called.
    """
    get_connection(using).on_commit(func)



import threading

import sleuth
from django.db import connection
from gcloudc.db import transaction
from gcloudc.db.backends.datastore import reset_context_cache

from . import TestCase
from .models import (
    TestFruit,
    TestUser,
)


class TransactionTests(TestCase):
    def test_on_commit_works(self):
        def increment():
            increment.x += 1

        increment.x = 0

        with transaction.atomic():
            connection.on_commit(increment)
            self.assertEqual(increment.x, 0)

        self.assertEqual(increment.x, 1)

        try:
            with transaction.atomic():
                connection.on_commit(increment)
                self.assertEqual(increment.x, 1)
                raise ValueError()
        except ValueError:
            pass

        self.assertEqual(increment.x, 1)  # Still the same

        with transaction.atomic():
            pass  # commit hook should have gone with rollback

        self.assertEqual(increment.x, 1)  # Still the same

    def test_get_or_create(self):
        """
            get_or_create uses Django's atomic decorator under the hood
            this can cause issues if called within a gcloudc atomic block
        """

        with transaction.atomic():
            user, created = TestUser.objects.get_or_create(username="foo")
            self.assertTrue(created)

            user, created = TestUser.objects.get_or_create(username="foo")
            self.assertFalse(created)

    def test_repeated_usage_in_a_loop(self):
        pk = TestUser.objects.create(username="foo").pk
        for i in range(4):
            with transaction.atomic(xg=True):
                TestUser.objects.get(pk=pk)
                continue

        with transaction.atomic(xg=True):
            TestUser.objects.get(pk=pk)

    def test_recursive_atomic(self):
        lst = []

        @transaction.atomic
        def txn():
            lst.append(True)
            if len(lst) == 3:
                return
            else:
                txn()

        txn()

    def test_recursive_non_atomic(self):
        lst = []

        @transaction.non_atomic
        def txn():
            lst.append(True)
            if len(lst) == 3:
                return
            else:
                txn()

        txn()

    def test_atomic_in_separate_thread(self):
        """ Regression test.  See #668. """
        @transaction.atomic
        def txn():
            return

        def target():
            txn()

        thread = threading.Thread(target=target)
        thread.start()
        thread.join()

    def test_non_atomic_in_separate_thread(self):
        """ Regression test.  See #668. """
        @transaction.non_atomic
        def txn():
            return

        def target():
            txn()

        thread = threading.Thread(target=target)
        thread.start()
        thread.join()

    def test_atomic_decorator(self):
        @transaction.atomic
        def txn():
            TestUser.objects.create(username="foo", field2="bar")
            self.assertTrue(transaction.in_atomic_block())
            raise ValueError()

        with self.assertRaises(ValueError):
            txn()

        self.assertEqual(0, TestUser.objects.count())

    def test_atomic_context_manager(self):
        with self.assertRaises(ValueError):
            with transaction.atomic():
                TestUser.objects.create(username="foo", field2="bar")
                raise ValueError()

        self.assertEqual(0, TestUser.objects.count())

    def test_non_atomic_context_manager(self):
        existing = TestUser.objects.create(username="existing", field2="exists", first_name="one", second_name="one")

        with transaction.atomic():
            self.assertTrue(transaction.in_atomic_block())

            user = TestUser.objects.create(username="foo", field2="bar", first_name="two", second_name="two")

            with transaction.non_atomic():
                # We're outside the transaction, so the user should not exist
                self.assertRaises(TestUser.DoesNotExist, TestUser.objects.get, pk=user.pk)
                self.assertFalse(transaction.in_atomic_block())

                with sleuth.watch("google.cloud.datastore.client.Client.get") as datastore_get:
                    TestUser.objects.get(pk=existing.pk)  # Should hit the cache, not the datastore

                self.assertFalse(datastore_get.called)

            with transaction.atomic(independent=True):
                user2 = TestUser.objects.create(username="foo2", field2="bar2", first_name="three", second_name="three")
                self.assertTrue(transaction.in_atomic_block())

                with transaction.non_atomic():
                    self.assertFalse(transaction.in_atomic_block())
                    self.assertRaises(TestUser.DoesNotExist, TestUser.objects.get, pk=user2.pk)

                    with transaction.non_atomic():
                        self.assertFalse(transaction.in_atomic_block())
                        self.assertRaises(TestUser.DoesNotExist, TestUser.objects.get, pk=user2.pk)

                        with sleuth.watch("google.cloud.datastore.client.Client.get") as datastore_get:
                            # Should hit the cache, not the Datastore
                            TestUser.objects.get(pk=existing.pk)
                            self.assertFalse(datastore_get.called)

                    self.assertFalse(transaction.in_atomic_block())
                    self.assertRaises(TestUser.DoesNotExist, TestUser.objects.get, pk=user2.pk)

                # Should hit the cache
                self.assertTrue(TestUser.objects.filter(pk=user2.pk).exists())
                self.assertTrue(transaction.in_atomic_block())

        self.assertFalse(transaction.in_atomic_block())

    def test_independent_argument(self):
        """
            We would get a XG error if the inner transaction was not independent
        """
        @transaction.atomic
        def txn1(_username, _fruit):
            @transaction.atomic(independent=True)
            def txn2(_fruit):
                TestFruit.objects.create(name=_fruit, color="pink")
                raise ValueError()

            TestUser.objects.create(username=_username)
            txn2(_fruit)

        with self.assertRaises(ValueError):
            txn1("test", "banana")

    def test_enable_cache_argument(self):
        user = TestUser.objects.create(username="randy", first_name="Randy")

        with sleuth.watch('gcloudc.db.backends.datastore.context.CacheDict.get') as cachedict_get:
            TestUser.objects.get(username="randy")
            self.assertEqual(cachedict_get.call_count, 1)

            with transaction.atomic(enable_cache=False):
                user.first_name = "Łukasz"
                user.save()

                non_cached = TestUser.objects.get(username="randy")
                # Result is not fetched from the cache
                self.assertEqual(non_cached.first_name, "Randy")
                self.assertEqual(cachedict_get.call_count, 1)

    def test_nested_decorator(self):
        # Nested decorator pattern we discovered can cause a connection_stack
        # underflow.

        @transaction.atomic
        def inner_txn():
            pass

        @transaction.atomic
        def outer_txn():
            inner_txn()

        # Calling inner_txn first puts it in a state which means it doesn't
        # then behave properly in a nested transaction.
        inner_txn()
        outer_txn()

    def test_atomic_context_manager_raises_error(self):
        """Test that the transaction.atomic context manager
        handles errors raised when starting a transaction.
        """
        with sleuth.detonate(
                "gcloudc.db.backends.datastore.transaction.NormalTransaction._enter",
                exception=UserWarning):
            with self.assertRaises(UserWarning):
                with transaction.atomic():
                    TestFruit.objects.create(name="Apple", color="Red")

        self.assertEqual(TestFruit.objects.all().count(), 0)

        # this should not throw a `RuntimeError`;
        # the `transaction.atomic()` context manager should handle errors & cleanup the transaction
        reset_context_cache()


class TransactionStateTests(TestCase):

    def test_has_already_read(self):
        apple = TestFruit.objects.create(name="Apple", color="Red")
        pear = TestFruit.objects.create(name="Pear", color="Green")

        with transaction.atomic(xg=True) as txn:
            self.assertFalse(txn.has_already_been_read(apple))
            self.assertFalse(txn.has_already_been_read(pear))

            apple.refresh_from_db()

            self.assertTrue(txn.has_already_been_read(apple))
            self.assertFalse(txn.has_already_been_read(pear))

            with transaction.atomic(xg=True) as txn:
                self.assertTrue(txn.has_already_been_read(apple))
                self.assertFalse(txn.has_already_been_read(pear))
                pear.refresh_from_db()
                self.assertTrue(txn.has_already_been_read(pear))

                with transaction.atomic(independent=True) as txn2:
                    self.assertFalse(txn2.has_already_been_read(apple))
                    self.assertFalse(txn2.has_already_been_read(pear))

    def test_refresh_if_unread(self):
        apple = TestFruit.objects.create(name="Apple", color="Red")

        with transaction.atomic() as txn:
            apple.color = "Pink"

            txn.refresh_if_unread(apple)

            self.assertEqual(apple.name, "Apple")

            apple.color = "Pink"

            # Already been read this transaction, don't read it again!
            txn.refresh_if_unread(apple)

            self.assertEqual(apple.color, "Pink")



import uuid
from django.db import models
from django.core.validators import EmailValidator
from gcloudc.db.models.fields.charfields import (
    CharField
)
from gcloudc.db.models.fields.computed import ComputedCollationField
from gcloudc.db.models.fields.related import RelatedSetField, RelatedListField, GenericRelationField
from gcloudc.db.models.fields.iterable import SetField, ListField


class BinaryFieldModel(models.Model):
    binary = models.BinaryField(null=True)


class ModelWithCharField(models.Model):
    char_field_with_max = CharField(max_length=10, default="", blank=True)
    char_field_without_max = CharField(default="", blank=True)
    char_field_as_email = CharField(max_length=100, validators=[EmailValidator(message="failed")], blank=True)


class TestUser(models.Model):
    """Basic model defintion for use in test cases."""

    username = models.CharField(max_length=32, unique=True)
    first_name = models.CharField(max_length=50)
    second_name = models.CharField(max_length=50)
    email = models.EmailField(blank=True, default="")
    field2 = models.CharField(max_length=32, blank=True, default="")

    last_login = models.DateTimeField(null=True, blank=True)

    def __str__(self):
        return self.username

    class Meta:
        unique_together = ("first_name", "second_name")


class TestUserTwo(models.Model):
    username = models.CharField(max_length=32, unique=True)


class BasicTestModel(models.Model):
    field1 = models.CharField(max_length=100)
    field2 = models.IntegerField(unique=True)


class MultiQueryModel(models.Model):
    field1 = models.IntegerField(null=True)
    field2 = models.CharField(max_length=64)


class ModelWithComputedCollationField(models.Model):
    """Test model for `ComputedCollationField`."""

    name = models.CharField(max_length=100)
    name_order = ComputedCollationField("name")


class PFPost(models.Model):
    content = models.TextField()
    authors = RelatedSetField('PFAuthor', related_name='posts')


class PFAuthor(models.Model):
    name = models.CharField(max_length=32)
    awards = RelatedSetField('PFAwards')


class PFAwards(models.Model):
    name = models.CharField(max_length=32)


class ISOther(models.Model):
    name = models.CharField(max_length=500)

    def __str__(self):
        return "%s:%s" % (self.pk, self.name)


class RelationWithoutReverse(models.Model):
    name = models.CharField(max_length=500)


class ISModel(models.Model):
    related_things = RelatedSetField(ISOther)
    related_list = RelatedListField(ISOther, related_name="ismodel_list")
    limted_related = RelatedSetField(RelationWithoutReverse, limit_choices_to={"name": "banana"}, related_name="+")
    children = RelatedSetField("self", related_name="+")


class GenericRelationModel(models.Model):
    relation_to_anything = GenericRelationField(null=True)
    unique_relation_to_anything = GenericRelationField(null=True, unique=True)


class IterableFieldsWithValidatorsModel(models.Model):
    set_field = SetField(models.CharField(max_length=100), min_length=2, max_length=3, blank=False)
    list_field = ListField(models.CharField(max_length=100), min_length=2, max_length=3, blank=False)
    related_set = RelatedSetField(ISOther, min_length=2, max_length=3, blank=False)
    related_list = RelatedListField(ISOther, related_name="iterable_list", min_length=2, max_length=3, blank=False)


class ModelDatabaseA(models.Model):
    set_of_bs = RelatedSetField("ModelDatabaseB", related_name="+")
    list_of_bs = RelatedListField("ModelDatabaseB", related_name="+")


class ModelDatabaseB(models.Model):
    test_database = "ns1"


class IterableRelatedModel(models.Model):
    related_set = RelatedListField(ISOther, related_name="+")
    related_list = RelatedListField(ISOther, related_name="+")


class RelationWithOverriddenDbTable(models.Model):
    class Meta:
        db_table = "bananarama"


class Post(models.Model):
    content = models.TextField()
    tags = RelatedSetField("Tag", related_name="posts")
    ordered_tags = RelatedListField("Tag")


class Tag(models.Model):
    name = models.CharField(max_length=64)


class RelatedCharFieldModel(models.Model):
    char_field = CharField(max_length=500)


class StringPkModel(models.Model):
    name = models.CharField(max_length=500, primary_key=True)


class IterableRelatedWithNonIntPkModel(models.Model):
    related_set = RelatedListField(StringPkModel, related_name="+")
    related_list = RelatedListField(StringPkModel, related_name="+")


class RelatedListFieldRemoveDuplicatesModel(models.Model):
    related_list_field = RelatedListField(RelatedCharFieldModel, remove_duplicates=True)


class ISStringReferenceModel(models.Model):
    related_things = RelatedSetField('ISOther')
    related_list = RelatedListField('ISOther', related_name="ismodel_list_string")
    limted_related = RelatedSetField('RelationWithoutReverse', limit_choices_to={'name': 'banana'}, related_name="+")
    children = RelatedSetField("self", related_name="+")


class TestFruit(models.Model):
    name = models.CharField(primary_key=True, max_length=32)
    origin = models.CharField(max_length=32, default="Unknown")
    color = models.CharField(max_length=100)
    is_mouldy = models.BooleanField(default=False)
    text_field = models.TextField(blank=True, default="")
    binary_field = models.BinaryField(blank=True)

    class Meta:
        ordering = ("color",)

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<TestFruit: name={}, color={}>".format(self.name, self.color)


class TransformTestModel(models.Model):
    field1 = models.CharField(max_length=255)
    field2 = models.CharField(max_length=255, unique=True)
    field3 = models.CharField(null=True, max_length=255)
    field4 = models.TextField()


class InheritedModel(TransformTestModel):
    pass


class Relation(models.Model):
    pass


class Related(models.Model):
    headline = models.CharField(max_length=500)
    relation = models.ForeignKey(Relation, on_delete=models.DO_NOTHING)


class IntegerModel(models.Model):
    integer_field = models.IntegerField()


class NullDate(models.Model):
    date = models.DateField(null=True, default=None)
    datetime = models.DateTimeField(null=True, default=None)
    time = models.TimeField(null=True, default=None)


class NullDateSet(models.Model):
    dates = RelatedSetField(NullDate, blank=True, unique=True)


class UniqueModel(models.Model):
    unique_field = models.CharField(max_length=100, unique=True)
    unique_combo_one = models.IntegerField(blank=True, default=0)
    unique_combo_two = models.CharField(max_length=100, blank=True, default="")

    unique_relation = models.ForeignKey('self', null=True, blank=True, unique=True, on_delete=models.DO_NOTHING)

    unique_set_field = SetField(models.CharField(max_length=500), unique=True)
    unique_list_field = ListField(models.CharField(max_length=500), unique=True)

    unique_together_list_field = ListField(models.IntegerField())

    class Meta:
        unique_together = [
            ("unique_combo_one", "unique_combo_two"),
            ("unique_together_list_field", "unique_combo_one")
        ]


class ModelWithNullableCharField(models.Model):
    field1 = models.CharField(max_length=500, null=True)
    some_id = models.IntegerField(default=0)


class DurationModel(models.Model):
    duration_field = models.DurationField()
    duration_field_nullable = models.DurationField(blank=True, null=True)


class ModelWithUniques(models.Model):
    name = models.CharField(max_length=64, unique=True)


class ModelWithUniquesOnForeignKey(models.Model):
    name = models.CharField(max_length=64, unique=True)
    related_name = models.ForeignKey(ModelWithUniques, unique=True, on_delete=models.DO_NOTHING)

    class Meta:
        unique_together = [("name", "related_name")]


class UniqueModelWithLongPK(models.Model):
    long_pk = models.CharField(max_length=500, primary_key=True)
    unique_field = models.IntegerField(unique=True)


class Zoo(models.Model):
    pass


class Enclosure(models.Model):
    zoo = models.ForeignKey(Zoo, on_delete=models.CASCADE)


class Animal(models.Model):
    enclosure = models.ForeignKey(Enclosure, on_delete=models.CASCADE)


class UUIDTestModel(models.Model):
    uuid = models.UUIDField(default=uuid.uuid4)


class SpecialIndexesModel(models.Model):
    name = models.CharField(max_length=255, primary_key=True)
    nickname = CharField(blank=True)
    sample_list = ListField(models.CharField)

    def __str__(self):
        return self.name


class Permission(models.Model):
    user = models.ForeignKey(TestUser, on_delete=models.CASCADE)
    perm = models.CharField(max_length=32)

    def __str__(self):
        return u"{0} for {1}".format(self.perm, self.user)

    class Meta:
        ordering = ('user__username', 'perm')


class SelfRelatedModel(models.Model):
    related = models.ForeignKey('self', blank=True, null=True, on_delete=models.SET_NULL)


class MultiTableParent(models.Model):
    parent_field = models.CharField(max_length=32)


class ModelWithDates(models.Model):
    start = models.DateField()
    end = models.DateField()


class MultiTableChildOne(MultiTableParent):
    child_one_field = models.CharField(max_length=32)


class MultiTableChildTwo(MultiTableParent):
    child_two_field = models.CharField(max_length=32)


class DateTimeModel(models.Model):
    datetime_field = models.DateTimeField(auto_now_add=True)
    date_field = models.DateField(auto_now_add=True)


class NonIndexedModel(models.Model):
    content = models.TextField()
    binary = models.BinaryField()


class NullableFieldModel(models.Model):
    nullable = models.IntegerField(null=True)


class Thing(models.Model):
    num = models.IntegerField(default=0)
